#
# for linux 
#
# export GUI_TYPE
# export BUILDDIR
# export LIBS
# export LDFLAGS
# export CDEFS
# export CFLAGS
# export INSTALLDIR

CC=gcc
CXX=g++
LD=g++

EXEFILE=mbs1

DATADIR=data

SRCDIR=src
SRCVM=$(SRCDIR)/vm
SRCFMGEN=$(SRCVM)/fmgen
SRCDEP=$(SRCVM)/mbs1
SRCVID=$(SRCDIR)/video
SRCVIDFFM=$(SRCDIR)/video/ffmpeg
SRCGUIAGAR=$(SRCDIR)/gui/agar
SRCGUIGTKX11=$(SRCDIR)/gui/gtk_x11
#ifeq ($(GUI_TYPE),GUI_TYPE_AGAR)
#	SRCGUI=$(SRCGUIAGAR)
#endif
#ifeq ($(GUI_TYPE),GUI_TYPE_GTK_X11)
#	SRCGUI=$(SRCGUIGTKX11)
#endif

RESDIR=res
SRCRES=$(SRCDIR)/$(RESDIR)/common

LOCALEDIR=locale
SRCLOCALE=$(LOCALEDIR)

EXE=$(BUILDDIR)/$(EXEFILE)

EMUOBJS=$(SRCDIR)/config.o \
	$(SRCDIR)/fileio.o \
	$(SRCDIR)/logging.o \
	$(SRCDIR)/emu.o \
	$(SRCDIR)/emumsg.o \
	$(SRCDIR)/sdl_input.o \
	$(SRCDIR)/sdl_input_keysym.o \
	$(SRCDIR)/sdl_screen.o \
	$(SRCDIR)/sdl_sound.o \
	$(SRCDIR)/sdl_timer.o \
	$(SRCDIR)/sdl_utils.o \
	$(SRCDIR)/ledbox.o \
	$(SRCDIR)/msgboard.o \
	$(SRCDIR)/sdl_socket.o \
	$(SRCDIR)/sdl_main.o \
	$(SRCDIR)/depend.o \
	$(SRCDIR)/clocale.o \
	$(SRCDIR)/parseopt.o \
	$(SRCDIR)/utility.o
#

VMOBJS= $(SRCVM)/event.o \
	$(SRCVM)/hd46505.o \
	$(SRCVM)/ym2203.o \
	$(SRCVM)/disk.o \
	$(SRCVM)/mc6809.o \
	$(SRCVM)/mc6809dasm.o

FMGENOBJS=$(SRCFMGEN)/fmgen.o \
	$(SRCFMGEN)/fmtimer.o \
	$(SRCFMGEN)/opm.o \
	$(SRCFMGEN)/opna.o \
	$(SRCFMGEN)/psg.o
#	$(SRCFMGEN)/file.o \
#

DEPOBJS=$(SRCDEP)/acia.o \
	$(SRCDEP)/mbs1.o \
	$(SRCDEP)/cmt.o \
	$(SRCDEP)/comm.o \
	$(SRCDEP)/display.o \
	$(SRCDEP)/kanji.o \
	$(SRCDEP)/keyboard.o \
	$(SRCDEP)/memory.o \
	$(SRCDEP)/pia.o \
	$(SRCDEP)/printer.o \
	$(SRCDEP)/sound.o \
	$(SRCDEP)/parsewav.o \
	$(SRCDEP)/mb8866.o \
	$(SRCDEP)/mc6843.o \
	$(SRCDEP)/floppy.o \
	$(SRCDEP)/via.o \
	$(SRCDEP)/psgc.o \
	$(SRCDEP)/psg9c.o \
	$(SRCDEP)/board.o \
	$(SRCDEP)/mouse.o \
	$(SRCDEP)/registers.o \
	$(SRCDEP)/keyrecord.o \
	$(SRCDEP)/timer.o 

GUIOBJSAGAR=$(SRCGUIAGAR)/ag_gui_base.o \
	$(SRCGUIAGAR)/ag_gui_config.o \
	$(SRCGUIAGAR)/ag_dlg.o \
	$(SRCGUIAGAR)/ag_file_dlg.o \
	$(SRCGUIAGAR)/ag_volume_dlg.o \
	$(SRCGUIAGAR)/ag_config_dlg.o \
	$(SRCGUIAGAR)/ag_keybind_dlg.o \
	$(SRCGUIAGAR)/ag_recvid_dlg.o \
	$(SRCGUIAGAR)/ag_dir_dlg.o \
	$(SRCGUIAGAR)/ag_gui.o \
	$(SRCGUIGTKX11)/gtk_x11_gui.o

GUIOBJSGTKX11=$(SRCGUIGTKX11)/gtk_x11_gui.o \
	$(SRCGUIGTKX11)/gtk_dialogbox.o \
	$(SRCGUIGTKX11)/gtk_configbox.o \
	$(SRCGUIGTKX11)/gtk_x11_key_trans.o \
	$(SRCGUIGTKX11)/gtk_keybindbox.o \
	$(SRCGUIGTKX11)/gtk_volumebox.o \
	$(SRCGUIGTKX11)/gtk_recvidbox.o \
	$(SRCGUIGTKX11)/gtk_filebox.o \
	$(SRCGUIGTKX11)/gtk_folderbox.o \
	$(SRCGUIGTKX11)/gtk_aboutbox.o

VIDOBJSFFM=$(SRCVIDFFM)/ffm_rec_video.o

VIDOBJS=$(VIDOBJSFFM) $(SRCVID)/rec_video.o

ifeq ($(GUI_TYPE),GUI_TYPE_AGAR)
	GUIOBJS=$(GUIOBJSAGAR)
endif
ifeq ($(GUI_TYPE),GUI_TYPE_GTK_X11)
	GUIOBJS=$(GUIOBJSGTKX11)
endif

GUIOBJS:=$(GUIOBJS) $(SRCDIR)/gui/gui_base.o \
	$(SRCDIR)/gui/gui_keybinddata.o

OBJS_BASE=$(DEPOBJS) $(VMOBJS) $(FMGENOBJS) $(EMUOBJS) $(GUIOBJS) $(VIDOBJS)
DEPS_BASE=$(OBJS_BASE:%.o=%.d)
OBJS=$(OBJS_BASE:%=$(BUILDDIR)/%)
DEPS=$(DEPS_BASE:%=$(BUILDDIR)/%)

MAKEFILEDEP=$(BUILDDIR)/Makefile.dep

#
#
#

all: exe

exe: $(BUILDDIR) $(EXE)

depend: $(MAKEFILEDEP)

$(MAKEFILEDEP): $(BUILDDIR) $(DEPS)
	cat $(DEPS) > $@

$(EXE): $(OBJS)
	$(LD) -o $@ $(OBJS) $(LIBS) $(LDFLAGS)

$(BUILDDIR)/%.o: %.cpp
	$(CXX) $(CDEFS) $(CFLAGS) -c $< -o $@

$(BUILDDIR)/%.d: %.cpp
	-$(CC) $(CDEFS) $(CFLAGS) -MM $< | sed 's#^.*\.o:#$@:#' | sed 's/\.d:/.o:/' > $@

$(BUILDDIR)/%.o: %.c
	$(CC) $(CDEFS) $(CFLAGS) -c $< -o $@

$(BUILDDIR)/%.d: %.c
	-$(CC) $(CDEFS) $(CFLAGS) -MM $< | sed 's#^.*\.o:#$@:#' | sed 's/\.d:/.o:/' > $@

install: exe
	mkdir -p $(INSTALLDIR)
	cp -p $(EXE) $(INSTALLDIR)
#	cp -p $(DATADIR)/?*.* $(INSTALLDIR)
	mkdir -p $(INSTALLDIR)/$(RESDIR)
	cp -p $(SRCRES)/*.* $(INSTALLDIR)/$(RESDIR)
	mkdir -p $(INSTALLDIR)/$(LOCALEDIR)
	for i in $(LOCALEDIR)/*/*; do if [ -d $$i ]; then \
		mkdir -p $(INSTALLDIR)/$$i; cp -p $$i/*.mo $(INSTALLDIR)/$$i; \
	fi; done

$(BUILDDIR):
	mkdir -p $(BUILDDIR)/$(SRCFMGEN)
	mkdir -p $(BUILDDIR)/$(SRCDEP)
	mkdir -p $(BUILDDIR)/$(SRCGUIAGAR)
	mkdir -p $(BUILDDIR)/$(SRCGUIGTKX11)
	mkdir -p $(BUILDDIR)/$(SRCVIDFFM)

clean:
	rm -rf $(BUILDDIR)

include $(MAKEFILEDEP)
