/** @file cocoa_configpanel.mm
 
 HITACHI MB-S1 Emulator 'EmuB-S1'
 Skelton for retropc emulator
 SDL edition + Cocoa GUI
 
 @author Sasaji
 @date   2015.04.30 -
 
 @brief [ config panel ]
 */

#import "cocoa_gui.h"
#import "cocoa_configpanel.h"
#import "../../config.h"
#import "../../emu.h"
#import "../../clocale.h"
#import "../../msgboard.h"
#import "cocoa_fontpanel.h"

extern EMU *emu;

@implementation CocoaConfigPanel
- (id)init
{
	const char *tab_labels[] = { _("Mode"),_("Screen"),_("Tape"),_("Network"),_("Memory"),NULL };
	const char *sys_labels[] = { _("A Mode (S1)"),_("B Mode (L3)"),NULL };
	const char *fdd_labels[] = {
		_("No FDD"),
		_("3inch compact FDD (for L3)"),
		_("5.25inch mini FDD (2D Type)"),
		_("5.25inch mini FDD (2HD Type)"),
	NULL };
	const char *io_labels[] = {
		_("5.25inch FDC  $FF00 - $FF14"),
		_("3inch FDC (for L3)  $FF18 - $FF20"),
		_("6voice PSG (for L3)  $FF30 - $FF33"),
		_("Ex LPT Port  $FF3C - $FF3F"),
		_("Ex COM Port  $FF40 - $FF41"),
		_("9voice PSG (for L3)  $FF70 - $FF7F"),
		_("KANJI ROM  $FF75 - $FF76"),
		_("Ex PSG (for S1)  $FFE6 - $FFE7"),
	NULL };
	const char *gl_use_labels[] = { _("No"),_("Yes(sync)"),_("Yes(async)"),NULL };
	const char *gl_filter_labels[] = { _("Nearest Neighbour"),_("Linear"),NULL };
	const char *led_pos_labels[] = { _("LeftTop"),_("RightTop"),_("LeftBottom"),_("RightBottom"),NULL };
	const char *skew_labels[] = { "-2","-1","0","1","2",NULL };
	const char *corr_labels[] = { _("COS Wave"),_("SIN Wave"),NULL };
	const char *rate_labels[] = { "11025", "22050", "44100", "48000", NULL };
	const char *bits_labels[] = { "8", "16", NULL };
	const char *comm_baud_labels[] = {
		_("S:300baud/F:1200baud"),
		_("S:600baud/F:2400baud"),
		_("S:1200baud/F:4800baud"),
		_("S:2400baud/F:9600baud"),
		NULL
	};
	const char *exram_labels[] = { _("None"), "64KB", "128KB", "256KB", "512KB", NULL };

	[super init];

	self.title = [NSString stringWithUTF8String:_("Configure")];
	[self setShowsResizeIndicator:FALSE];
	
	CocoaView *view = [self contentView];
	NSRect re_tab;
	int top = view.margin;
	int left = view.margin;
	
	re_tab = [view makeRect:left:top:300:200];
	
	CocoaTabView *tabView = [CocoaTabView create:re_tab tabs:tab_labels];
	[view addSubview:tabView];

	NSRect re;
	NSTabViewItem *tab;
	CocoaView *tab_view;
	NSRect re_box;
	CocoaBox *box;
	CocoaView *box_view;
	CocoaLabel *lbl;
	CocoaButton *btn;

	int bleft = view.margin;
	int btop = view.margin;
	int i;
	int center_left = 0;
	int middle_top = 0;

	// Mode tab
	tab = [tabView tabViewItemAtIndex:0];
	tab_view = (CocoaView *)[tab view];

	//
	re_box = [tab_view makeRect:bleft:btop:260:60];
	box = [CocoaBox create:re_box title:_("System Mode (*)")];
	box_view = [box contentView];
	
	re = [box_view makeRect:bleft:btop+2:16:24];
	lbl = [CocoaLabel create:re title:(config.sys_mode & 1) == 1 ? ">" : " "];
	[box_view addSubview:lbl];
	re = [box_view addHeight:re:16:24];
	lbl = [CocoaLabel create:re title:(config.sys_mode & 1) == 0 ? ">" : " "];
	[box_view addSubview:lbl];
	
	re = [box_view remakeTop:re:btop];
	re = [box_view addWidth:re:240:48];
	radSysMode = [CocoaRadioGroup create:re rows:2 cols:1 titles:sys_labels action:nil selidx:(1 - (emu->get_parami(EMU::ParamSysMode) & 1))];
	[box_view addSubview:radSysMode];
	
	re_box = [box adjustFrameSize:re_box];
	if (center_left < box_view.max_right) center_left = box_view.max_right;
	[tab_view addSubview:box];

	//
	re_box = [tab_view addHeight:re_box:260:200];
	box = [CocoaBox create:re_box title:_("FDD Type (*)")];
	box_view = [box contentView];
	
	re = [box_view makeRect:bleft:btop+2:16:24];
	lbl = [CocoaLabel create:re title:(config.fdd_type == 0) ? ">" : " "];
	[box_view addSubview:lbl];
	re = [box_view addHeight:re:16:24];
	lbl = [CocoaLabel create:re title:(config.fdd_type == 1) ? ">" : " "];
	[box_view addSubview:lbl];
	re = [box_view addHeight:re:16:24];
	lbl = [CocoaLabel create:re title:(config.fdd_type == 2) ? ">" : " "];
	[box_view addSubview:lbl];
	re = [box_view addHeight:re:16:24];
	lbl = [CocoaLabel create:re title:(config.fdd_type == 3) ? ">" : " "];
	[box_view addSubview:lbl];
	
	re = [box_view remakeTop:re:btop];
	re = [box_view addWidth:re:240:96];
	radFddType = [CocoaRadioGroup create:re rows:4 cols:1 titles:fdd_labels action:@selector(selectFddType:) selidx:emu->get_parami(EMU::ParamFddType)];
	[box_view addSubview:radFddType];

	re_box = [box adjustFrameSize:re_box];
	if (center_left < box_view.max_right) center_left = box_view.max_right;
	middle_top = [CocoaView bottomPos:re_box];
	[tab_view addSubview:box];

	//
	re_box = [tab_view makeRect:center_left+bleft:btop:320:200];
	box = [CocoaBox create:re_box title:_("I/O Port Address (*)")];
	box_view = [box contentView];

	re = [box_view makeRect:bleft:btop:0:0];
	for(i=0; i<8; i++) {
		re = [box_view remakeLeft:re:bleft];
		re = [box_view addHeight:re:16:24];
		lbl = [CocoaLabel create:re title:(config.io_port & (1 << i)) ? ">" : " "];
		[box_view addSubview:lbl];
		
		re = [box_view addWidth:re:270:24];
		chkIOPort[i] = [CocoaCheckBox create:re title:io_labels[i] index:i action:@selector(selectIO:) value:(emu->get_parami(EMU::ParamIOPort) & (1 << i))];
		[box_view addSubview:chkIOPort[i]];
		[self selectIO:chkIOPort[i]];
	}
	re_box = [box adjustFrameSize:re_box];
	if (middle_top < [CocoaView bottomPos:re_box]) middle_top = [CocoaView bottomPos:re_box];
	[tab_view addSubview:box];

	// mount fdd
	re_box = [tab_view makeRect:bleft:middle_top:200:24];
	lbl = [CocoaLabel createWithFitW:&re_box title:_("When start up, mount a disk in FDD:")];
	[tab_view addSubview:lbl];
	for(i=0; i<MAX_DRIVE; i++) {
		char name[4];
		sprintf(name, "%d ", i);
		re_box = [tab_view addWidth:re_box:32:24];
		chkFddMount[i] = [CocoaCheckBox create:re_box title:name action:nil value:(config.mount_fdd & (1 << i))];
		[tab_view addSubview:chkFddMount[i]];
	}

	//
	re_box = [tab_view addHeight:re_box:260:32];
	re_box = [tab_view remakeLeft:re_box:bleft];
	lbl = [CocoaLabel createWithFitW:&re_box title:_("* Need restart program or \"PowerOn\".")];
	[tab_view addSubview:lbl];


	// Screen tab
	tab = [tabView tabViewItemAtIndex:1];
	tab_view = (CocoaView *)[tab view];
	
	//
	re_box = [tab_view makeRect:bleft:btop:260:100];
	box = [CocoaBox create:re_box title:_("OpenGL")];
	box_view = [box contentView];

	re = [box_view makeRect:bleft:btop:80:24];
	lbl = [CocoaLabel create:re title:_("Use (*)")];
	[box_view addSubview:lbl];
	
	re = [box_view addWidth:re:180:24];
	popUseOpenGL = [CocoaPopUpButton create:re items:gl_use_labels action:nil selidx:config.use_opengl];
	[box_view addSubview:popUseOpenGL];
	
	re = [box_view remakeLeft:re:bleft];
	re = [box_view addHeight:re:80:24];
	lbl = [CocoaLabel create:re title:_("Filter Type")];
	[box_view addSubview:lbl];
	
	re = [box_view addWidth:re:180:24];
	popGLFilter = [CocoaPopUpButton create:re items:gl_filter_labels action:nil selidx:config.gl_filter_type];
	[box_view addSubview:popGLFilter];

	re_box = [box adjustFrameSize:re_box];
	[tab_view addSubview:box];

	//
	re_box = [tab_view addWidth:re_box:240:100];
	box = [CocoaBox create:re_box title:_("CRTC")];
	box_view = [box contentView];

	re = [box_view makeRect:bleft:btop:180:24];
	lbl = [CocoaLabel create:re title:_("Disptmg Skew")];
	[box_view addSubview:lbl];

	re = [box_view addWidth:re:60:24];
	popDisptmgSkew = [CocoaPopUpButton create:re items:skew_labels action:nil selidx:(config.disptmg_skew + 2)];
	[box_view addSubview:popDisptmgSkew];

	re = [box_view remakeLeft:re:bleft];
	re = [box_view addHeight:re:180:24];
	lbl = [CocoaLabel create:re title:_("Curdisp Skew (L3 Only)")];
	[box_view addSubview:lbl];

	re = [box_view addWidth:re:60:24];
	popCurdispSkew = [CocoaPopUpButton create:re items:skew_labels action:nil selidx:(config.curdisp_skew + 2)];
	[box_view addSubview:popCurdispSkew];

	re_box = [box adjustFrameSize:re_box];
	[tab_view addSubview:box];
	
	//
	re_box = [tab_view remakeLeft:re_box:bleft];
	re_box = [tab_view addHeight:re_box:40:24];
	lbl = [CocoaLabel create:re_box title:_("LED")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:60:24];
	lbl = [CocoaLabel create:re_box title:_("Position")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:160:24];
	popLEDPosition = [CocoaPopUpButton create:re_box items:led_pos_labels action:nil selidx:config.led_pos];
	[tab_view addSubview:popLEDPosition];

	//
	re_box = [tab_view remakeLeft:re_box:bleft];
	re_box = [tab_view addHeight:re_box:140:26];
	lbl = [CocoaLabel create:re_box title:_("Snapshot Path")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:360:26];
	txtSnapPath = [CocoaTextField create:re_box text:config.snapshot_path action:nil];
	[tab_view addSubview:txtSnapPath];
	
	re_box = [tab_view addWidth:re_box:100:26];
	btn = [CocoaButton create:re_box title:_("Folder...") action:@selector(showFolderPanel:)];
	[btn setRelatedObject:txtSnapPath];
	[tab_view addSubview:btn];

	//
	re_box = [tab_view remakeLeft:re_box:bleft];
	re_box = [tab_view addHeight:re_box:140:26];
	lbl = [CocoaLabel create:re_box title:_("Font Path")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:360:26];
	txtFontPath = [CocoaTextField create:re_box text:config.font_path action:nil];
	[tab_view addSubview:txtFontPath];
	
	re_box = [tab_view addWidth:re_box:100:26];
	btn = [CocoaButton create:re_box title:_("Folder...") action:@selector(showFolderPanel:)];
	[btn setRelatedObject:txtFontPath];
	[tab_view addSubview:btn];

	//
	re_box = [tab_view remakeLeft:re_box:bleft];
	re_box = [tab_view addHeight:re_box:140:24];
	lbl = [CocoaLabel create:re_box title:_("Message Font")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:200:24];
	txtMsgFontName = [CocoaTextField create:re_box text:config.msgboard_msg_fontname action:nil];
	[tab_view addSubview:txtMsgFontName];
	
	re_box = [tab_view addWidth:re_box:100:24];
	lbl = [CocoaLabel createWithFit:&re_box title:_(" Size")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:100:24];
	txtMsgFontSize = [CocoaTextField create:re_box num:config.msgboard_msg_fontsize action:nil];
	[tab_view addSubview:txtMsgFontSize];

	re_box = [tab_view addWidth:re_box:100:26];
	btn = [CocoaButton create:re_box title:_("File...") action:@selector(showFilePanel:)];
	[btn setRelatedObject:txtMsgFontName];
	[tab_view addSubview:btn];

	//
	re_box = [tab_view remakeLeft:re_box:bleft];
	re_box = [tab_view addHeight:re_box:140:24];
	lbl = [CocoaLabel create:re_box title:_("Info Font")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:200:24];
	txtInfoFontName = [CocoaTextField create:re_box text:config.msgboard_info_fontname action:nil];
	[tab_view addSubview:txtInfoFontName];
	
	re_box = [tab_view addWidth:re_box:100:24];
	lbl = [CocoaLabel createWithFitW:&re_box title:_(" Size")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:100:24];
	txtInfoFontSize = [CocoaTextField create:re_box num:config.msgboard_info_fontsize action:nil];
	[tab_view addSubview:txtInfoFontSize];

	re_box = [tab_view addWidth:re_box:100:26];
	btn = [CocoaButton create:re_box title:_("File...") action:@selector(showFilePanel:)];
	[btn setRelatedObject:txtInfoFontName];
	[tab_view addSubview:btn];

	//
	re_box = [tab_view remakeLeft:re_box:bleft];
	re_box = [tab_view addHeight:re_box:300:24];
	lbl = [CocoaLabel create:re_box title:_("* Need restart program.")];
	[tab_view addSubview:lbl];

	// Tape tab
	tab = [tabView tabViewItemAtIndex:2];
	tab_view = (CocoaView *)[tab view];

	//
	re_box = [tab_view makeRect:bleft:btop:300:200];
	box = [CocoaBox create:re_box title:_("Load Wav File")];
	box_view = [box contentView];

	re = [box_view makeRect:bleft:btop:260:24];
	chkReverseWave = [CocoaCheckBox create:re title:_("Reverse Wave") action:nil value:config.wav_reverse];
	[box_view addSubview:chkReverseWave];

	re = [box_view addHeight:re:260:24];
	chkHalfWave = [CocoaCheckBox create:re title:_("Half Wave") action:nil value:config.wav_half];
	[box_view addSubview:chkHalfWave];

	re = [box_view addHeight:re:260:24];
	chkCorrectWave = [CocoaCheckBox create:re title:_("Correct") action:@selector(selectCorrect:) value:config.wav_correct];
	[box_view addSubview:chkCorrectWave];

	re = [box_view addHeight:re:260:24];
	radCorrect = [CocoaRadioGroup create:re rows:1 cols:2 titles:corr_labels action:nil selidx:config.wav_correct_type ];
	[box_view addSubview:radCorrect];
	[self selectCorrect:chkCorrectWave];

	re_box = [box adjustFrameSize:re_box];
	[tab_view addSubview:box];

	//
	re_box = [tab_view addWidth:re_box:300:200];
	box = [CocoaBox create:re_box title:_("Save Wav File")];
	box_view = [box contentView];
	
	re = [box_view makeRect:bleft:btop:120:24];
	lbl = [CocoaLabel create:re title:_("Sample Rate")];
	[box_view addSubview:lbl];
	
	re = [box_view addWidth:re:120:24];
	popSampleRate = [CocoaPopUpButton create:re items:rate_labels action:nil selidx:config.wav_sample_rate];
	[box_view addSubview:popSampleRate];
	
	re = [box_view remakeLeft:re:bleft];
	re = [box_view addHeight:re:120:24];
	lbl = [CocoaLabel create:re title:_("Sample Bits")];
	[box_view addSubview:lbl];
	
	re = [box_view addWidth:re:120:24];
	popSampleBits = [CocoaPopUpButton create:re items:bits_labels action:nil selidx:config.wav_sample_bits];
	[box_view addSubview:popSampleBits];
	
	re_box = [box adjustFrameSize:re_box];
	[tab_view addSubview:box];

	// Network tab
	tab = [tabView tabViewItemAtIndex:3];
	tab_view = (CocoaView *)[tab view];
	
	re_box = [tab_view makeRect:bleft:btop:0:0];
	for(i=0; i<MAX_PRINTER; i++) {
		char name[128];
		sprintf(name, _("LPT%d Hostname"), i);
		re_box = [tab_view remakeLeft:re_box:bleft];
		re_box = [tab_view addHeight:re_box:120:24];
		lbl = [CocoaLabel create:re_box title:name];
		[tab_view addSubview:lbl];
		
		re_box = [tab_view addWidth:re_box:140:24];
		txtLPTHost[i] = [CocoaTextField create:re_box text:config.printer_server_host[i] action:nil];
		[tab_view addSubview:txtLPTHost[i]];
		
		re_box = [tab_view addWidth:re_box:60:24];
		lbl = [CocoaLabel createWithFitW:&re_box title:_(" Port") align:NSRightTextAlignment];
		[tab_view addSubview:lbl];
		
		re_box = [tab_view addWidth:re_box:80:24];
		txtLPTPort[i] = [CocoaTextField create:re_box num:config.printer_server_port[i] action:nil align:NSRightTextAlignment];
		[tab_view addSubview:txtLPTPort[i]];
	}
	for(i=0; i<MAX_COMM; i++) {
		char name[128];
		sprintf(name, _("COM%d Hostname"), i);
		re_box = [tab_view remakeLeft:re_box:bleft];
		re_box = [tab_view addHeight:re_box:120:24];
		lbl = [CocoaLabel create:re_box title:name];
		[tab_view addSubview:lbl];
		
		re_box = [tab_view addWidth:re_box:140:24];
		txtCOMHost[i] = [CocoaTextField create:re_box text:config.comm_server_host[i] action:nil];
		[tab_view addSubview:txtCOMHost[i]];
		
		re_box = [tab_view addWidth:re_box:60:24];
		lbl = [CocoaLabel createWithFitW:&re_box title:_(" Port") align:NSRightTextAlignment];
		[tab_view addSubview:lbl];
		
		re_box = [tab_view addWidth:re_box:80:24];
		txtCOMPort[i] = [CocoaTextField create:re_box num:config.comm_server_port[i] action:nil align:NSRightTextAlignment];
		[tab_view addSubview:txtCOMPort[i]];

		re_box = [tab_view addWidth:re_box:180:24];
		popCOMDipswitch[i] = [CocoaPopUpButton create:re_box items:comm_baud_labels action:nil selidx:(config.comm_dipswitch[i]-1)];
		[tab_view addSubview:popCOMDipswitch[i]];
	}

	// Memory tab
	tab = [tabView tabViewItemAtIndex:4];
	tab_view = (CocoaView *)[tab view];

	re_box = [tab_view makeRect:bleft:btop:100:26];
	lbl = [CocoaLabel create:re_box title:_("ROM Path")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:380:26];
	txtROMPath = [CocoaTextField create:re_box text:config.rom_path action:nil];
	[tab_view addSubview:txtROMPath];
	
	re_box = [tab_view addWidth:re_box:100:26];
	btn = [CocoaButton create:re_box title:_("Folder...") action:@selector(showFolderPanel:)];
	[btn setRelatedObject:txtROMPath];
	[tab_view addSubview:btn];

	
	re_box = [tab_view remakeLeft:re_box:bleft];
	re_box = [tab_view addHeight:re_box:120:26];
	lbl = [CocoaLabel create:re_box title:_("Extended RAM")];
	[tab_view addSubview:lbl];
	
	re_box = [tab_view addWidth:re_box:120:26];
	popExtRam = [CocoaPopUpButton create:re_box items:exram_labels action:nil selidx:emu->get_parami(EMU::ParamExMemNum)];
	[tab_view addSubview:popExtRam];

	re_box = [tab_view addWidth:re_box:120:26];
	if (0 <= config.exram_size_num && config.exram_size_num <= 4) {
		char str[128];
		strcpy(str, _(" (Now "));
		strcat(str, exram_labels[config.exram_size_num]);
		strcat(str, ")");
		lbl = [CocoaLabel create:re_box title:str];
		[tab_view addSubview:lbl];
	}
	
	re_box = [tab_view remakeLeft:re_box:bleft];
	re_box = [tab_view addHeight:re_box:300:24];
	lbl = [CocoaLabel create:re_box title:_("* Need restart program or \"PowerOn\".")];
	[tab_view addSubview:lbl];

	re_tab = [tabView adjustFrameSize:re_tab];

	// button
	
	re = [view makeRect:left+re_tab.size.width-120:top+re_tab.size.height:120:32];
	CocoaButton *btnCancel = [CocoaButton create:re title:_("Cancel") action:@selector(dialogCancel:)];
	[view addSubview:btnCancel];

	re = [view addPosition:re:-120-view.margin:0:120:32];
	CocoaButton *btnOK = [CocoaButton create:re title:_("OK") action:@selector(dialogOk:)];
	[view addSubview:btnOK];
	
	
	[self adjustFrame];
	
	[self selectFddType:radFddType];
	
	return self;
}

- (NSInteger)runModal
{
	return [NSApp runModalForWindow:self];
}

- (void)close
{
	[NSApp stopModalWithCode:NSCancelButton];
	[super close];
}

- (void)dialogOk:(id)sender
{
	int i;
	int valuel;

	// set data
	emu->set_parami(EMU::ParamSysMode,(1-(int)[radSysMode selectedRow]) | (config.sys_mode & ~1));
	emu->set_parami(EMU::ParamFddType,(int)[radFddType selectedRow]);

	for(i=0;i<MAX_DRIVE;i++) {
		config.mount_fdd = (([chkFddMount[i] state] == NSOnState) ? config.mount_fdd | (1 << i) : config.mount_fdd & ~(1 << i));
	}

	for(i=0;i<8;i++) {
		int val = emu->get_parami(EMU::ParamIOPort);
		if ([chkIOPort[i] state] == NSOnState) {
			val |= (1 << i);
		} else {
			val &= ~(1 << i);
		}
		emu->set_parami(EMU::ParamIOPort, val);
	}

	
	emu->set_parami(EMU::ParamUseOpenGL, (int)[popUseOpenGL indexOfSelectedItem]);
	config.gl_filter_type = [popGLFilter indexOfSelectedItem];
	
	config.led_pos = [popLEDPosition indexOfSelectedItem];
	
	config.curdisp_skew = [popCurdispSkew indexOfSelectedItem] - 2;
	config.disptmg_skew = [popDisptmgSkew indexOfSelectedItem] - 2;
	
	memset(config.snapshot_path, 0, sizeof(config.snapshot_path));
	strncpy(config.snapshot_path, [[txtSnapPath stringValue] UTF8String], _MAX_PATH-1);
	UTILITY::add_path_separator(config.snapshot_path);
	memset(config.font_path, 0, sizeof(config.font_path));
	strncpy(config.font_path, [[txtFontPath stringValue] UTF8String], _MAX_PATH-1);
	UTILITY::add_path_separator(config.font_path);
	
	memset(config.msgboard_msg_fontname, 0, sizeof(config.msgboard_msg_fontname));
	strncpy(config.msgboard_msg_fontname, [[txtMsgFontName stringValue] UTF8String], _MAX_PATH-1);
	memset(config.msgboard_info_fontname, 0, sizeof(config.msgboard_info_fontname));
	strncpy(config.msgboard_info_fontname, [[txtInfoFontName stringValue] UTF8String], _MAX_PATH-1);
	valuel = [txtMsgFontSize intValue];
	if (1 <= valuel && valuel <= 60) {
		config.msgboard_msg_fontsize = (int)valuel;
	}
	valuel = [txtInfoFontSize intValue];
	if (1 <= valuel && valuel <= 60) {
		config.msgboard_info_fontsize = (int)valuel;
	}

	config.wav_reverse = ([chkReverseWave state] == NSOnState);
	config.wav_half = ([chkHalfWave state] == NSOnState);
	config.wav_correct = ([chkCorrectWave state] == NSOnState);
	config.wav_correct_type = [radCorrect selectedColumn];

	config.wav_sample_rate = [popSampleRate indexOfSelectedItem];
	config.wav_sample_bits = [popSampleBits indexOfSelectedItem];
	
	for(i=0;i<MAX_PRINTER;i++) {
		memset(config.printer_server_host[i], 0, sizeof(config.printer_server_host[i]));
		strncpy(config.printer_server_host[i],[[txtLPTHost[i] stringValue] UTF8String], _MAX_PATH-1);
		valuel = [txtLPTPort[i] intValue];
		if (0 <= valuel && valuel <= 65535) {
			config.printer_server_port[i] = (int)valuel;
		}
	}
	for(i=0;i<MAX_COMM;i++) {
		memset(config.comm_server_host[i], 0, sizeof(config.comm_server_host[i]));
		strncpy(config.comm_server_host[i],[[txtCOMHost[i] stringValue] UTF8String], _MAX_PATH-1);
		valuel = [txtCOMPort[i] intValue];
		if (0 <= valuel && valuel <= 65535) {
			config.comm_server_port[i] = (int)valuel;
		}
		config.comm_dipswitch[i] = (int)[popCOMDipswitch[i] indexOfSelectedItem] + 1;
	}
	
	memset(config.rom_path, 0, sizeof(config.rom_path));
	strncpy(config.rom_path, [[txtROMPath stringValue] UTF8String], _MAX_PATH-1);
	UTILITY::add_path_separator(config.rom_path);

	emu->set_parami(EMU::ParamExMemNum, (int)[popExtRam indexOfSelectedItem]);

	// set message font
	MsgBoard *msgboard = emu->get_msgboard();
	if (msgboard) {
		msgboard->SetFont();
	}

	save_config();
	emu->change_pos_led(config.led_pos);
	emu->change_opengl_attr();

    // OK button is pushed
	[NSApp stopModalWithCode:NSOKButton];
	[super close];
}

- (void)dialogCancel:(id)sender
{
    // Cancel button is pushed
	[self close];
}

- (void)showFolderPanel:(id)sender
{
	NSInteger result;

	CocoaTextField *text = [sender relatedObject];
	if (text == nil) return;
	
	NSOpenPanel *panel = [NSOpenPanel openPanel];

	// cannot select file
	[panel setCanChooseFiles:NO];
	[panel setCanChooseDirectories:YES];
	
	// set current folder
	[panel setDirectoryURL:[NSURL fileURLWithPath:[text stringValue]]];
	
	// Display modal dialog
	result = [panel runModal];
	
	if(result == NSOKButton) {
		// get file path (use NSURL)
		NSURL *filePath = [panel URL];

		[text setStringValue:[filePath path]];
	}
}

- (void)showFilePanel:(id)sender
{
	NSInteger result;
	
	CocoaTextField *text = [sender relatedObject];
	if (text == nil) return;
	
	NSOpenPanel *panel = [NSOpenPanel openPanel];
	
	// can select file
	[panel setCanChooseFiles:YES];
	[panel setCanChooseDirectories:NO];
	
	// Display modal dialog
	result = [panel runModal];
	
	if(result == NSOKButton) {
		// get file path (use NSURL)
		NSURL *filePath = [panel URL];
		NSArray *arr = [filePath pathComponents];
		[text setStringValue:[arr lastObject]];
	}
}

- (void)showFontPanel:(id)sender
{
	NSInteger result;
	
	CocoaObjectStructure *obj = [sender relatedObject];
	if (obj == nil) return;
	CocoaTextField *fname = [obj obj1];
	if (fname == nil) return;
	CocoaTextField *fsize = [obj obj2];
	if (fsize == nil) return;
	
	// set font
	NSFont *font = [NSFont fontWithName:[fname stringValue] size:[fsize integerValue]];

	CocoaFontPanel *panel = [[CocoaFontPanel alloc] init];
	
	[panel setPanelFont:font isMultiple:NO];
	
	// Display modal dialog
	result = [panel runModal];

	if(result == NSOKButton) {
	}
//	[[NSFontManager sharedFontManager]orderFrontFontPanel:self];
}

- (void)changeFont:(id)sender
{

}

- (void)selectFddType:(CocoaRadioGroup *)sender
{
	int num = (int)[sender selectedRow];
	switch(num) {
		case 0:
			// No FDD
			[chkIOPort[0] setState:NSOffState];
			[chkIOPort[1] setState:NSOffState];
			[chkIOPort[0] setEnabled:YES];
			[chkIOPort[1] setEnabled:YES];
			break;
		case 1:
			// 3inch FDD
			[chkIOPort[0] setState:NSOffState];
			[chkIOPort[1] setState:NSOnState];
			[chkIOPort[0] setEnabled:NO];
			[chkIOPort[1] setEnabled:NO];
			break;
		case 2:
		case 3:
			// 5inch FDD
			[chkIOPort[0] setState:NSOnState];
			[chkIOPort[1] setState:NSOffState];
			[chkIOPort[0] setEnabled:NO];
			[chkIOPort[1] setEnabled:NO];
			break;
		default:
			break;
	}
}

- (void)selectIO:(CocoaCheckBox *)sender
{
	int num = [sender index];
	switch(num) {
		case 0:
			// 5inch FDD
			if ([sender state] == NSOnState) {
				[chkIOPort[1] setState:NSOffState];
			}
			break;
		case 1:
			// 3inch FDD
			if ([sender state] == NSOnState) {
				[chkIOPort[0] setState:NSOffState];
			}
			break;
		case 5:
			// 9psg
			if ([sender state] == NSOnState) {
				[chkIOPort[6] setState:NSOffState];
			}
			break;
		case 6:
			// Kanji
			if ([sender state] == NSOnState) {
				[chkIOPort[5] setState:NSOffState];
			}
			break;
		default:
			break;
	}
}

- (void)selectCorrect:(CocoaCheckBox *)sender
{
//	if ([sender state] == NSOnState) {
//		[radCorrect setEnabled:YES];
//	} else {
//		[radCorrect setEnabled:NO];
//	}
}

@end
