/** @file gtk_configbox.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2015.05.10 -

	@brief [ config box ]
*/

#include "gtk_configbox.h"
#include "gtk_folderbox.h"
#include "../../emu.h"
#include "../../config.h"
#include "../../clocale.h"
#include "../../msgboard.h"
#include "../../utility.h"

extern EMU *emu;

namespace GUI_GTK_X11
{

ConfigBox::ConfigBox(GUI *new_gui) : DialogBox(new_gui)
{
	memset(chkIO, 0, sizeof(chkIO));
}

ConfigBox::~ConfigBox()
{
}

bool ConfigBox::Show(GtkWidget *parent_window)
{
	DialogBox::Show(parent_window);

	if (dialog) return true;

	int flags = GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT;
	dialog = gtk_dialog_new_with_buttons(_("Configure"), GTK_WINDOW(window),
		(GtkDialogFlags)flags,
		_("OK"), GTK_RESPONSE_ACCEPT,
		_("Cancel"), GTK_RESPONSE_REJECT,
		NULL);
	GtkWidget *cont = gtk_dialog_get_content_area(GTK_DIALOG(dialog));
	GtkWidget *nb;
	GtkWidget *vboxall;
	GtkWidget *hboxall;
	GtkWidget *vboxl;
	GtkWidget *vboxr;
	GtkWidget *hbox;
	GtkWidget *vbox;
//	GtkWidget *frm;
//	GtkWidget *lbl;
//	GtkWidget *btn;

	char buf[128];

	// create notebook tab
	nb = create_notebook(cont);

	// 0 Mode

	vboxall = create_vbox(NULL);
	add_note(nb, vboxall, _("Mode"));

	hboxall = create_hbox(vboxall);
	vboxl = create_vbox(hboxall);

	create_frame(vboxl, _("DIP Switch (*)"), &vbox, &hbox);
	create_label(hbox, (config.dipswitch & 2) ? ">" : " ");
	chkMODE = create_check_box(hbox, _("MODE Switch"), config.dipswitch & 2);

	int fdd_type = emu->get_parami(EMU::ParamFddType);
	const char *fdd_type_label[] = {
		_("No FDD"),
		_("Use 3inch compact FDD"),
		_("Use 5.25inch mini FDD"),
		NULL
	};
	create_frame(vboxl, _("FDD Type (*)"), &vbox);
	for(int i=0; i<3; i++) {
		hbox = create_hbox(vbox);
		create_label(hbox, config.fdd_type == i ? ">" : " ");
		radFDD[i] = create_radio_box(hbox,radFDD,fdd_type_label[i],i,fdd_type, G_CALLBACK(OnChangedFDD));
	}

	vboxr = create_vbox(hboxall);

	int io_port = emu->get_parami(EMU::ParamIOPort);
	const char *io_port_label[] = {
		_("5.25inch FDC  $FF00 - $FF04"),
		_("3inch FDC  $FF18 - $FF20"),
		_("6voice PSG  $FF30 - $FF33"),
		_("Ex LPT Port  $FF3C - $FF3F"),
		_("Ex COM Port  $FF40 - $FF41"),
		_("9voice PSG  $FF70 - $FF7F"),
		_("KANJI ROM  $FF75 - $FF76"),
		NULL
	};
	create_frame(vboxr, _("I/O Port Address (*)"), &vbox);
	for(int i=0; i<7; i++) {
		hbox = create_hbox(vbox);
		create_label(hbox, config.io_port & (1 << i) ? ">" : " ");
		chkIO[i] = create_check_box(hbox,io_port_label[i],io_port & (1 << i),i,G_CALLBACK(OnChangedIO));
	}
	ChangeFDDType(get_radio_state_idx(radFDD,3));

	hbox = create_hbox(vboxall);
	create_label(hbox, _("* Need restart program or \"PowerOn\"."));

	// 1 screen

	vboxall = create_vbox(NULL);
	add_note(nb, vboxall, _("Screen"));

	hboxall = create_hbox(vboxall);

	const char *gl_use_label[] = {
		_("No"),
		_("Yes(sync)"),
		_("Yes(async)"),
		NULL
	};
	const char *gl_filter_label[] = {
		_("Nearest Neighbour"),
		_("Linear"),
		NULL
	};
	create_frame(hboxall, _("OpenGL"), &vbox, &hbox);
	comUseOpenGL = create_combo_box(hbox,_("Use (*)"),gl_use_label,config.use_opengl);
	hbox = create_hbox(vbox);
	comGLFilter = create_combo_box(hbox,_("Filter Type"),gl_filter_label,config.gl_filter_type);

	const char *skew_label[] = {
		"0",
		"1",
		"2",
		NULL
	};
	create_frame(hboxall, _("CRTC"), &vbox, &hbox);
	comDispSkew = create_combo_box(hbox,_("Disptmg Skew"),skew_label,config.disptmg_skew);
	hbox = create_hbox(vbox);
	comCurdSkew = create_combo_box(hbox,_("Curdisp Skew"),skew_label,config.curdisp_skew);

	const char *led_pos_label[] = {
		_("LeftTop"),
		_("RightTop"),
		_("LeftBottom"),
		_("RightBottom"),
		NULL
	};
	hbox = create_hbox(vboxall);
	create_label(hbox, _("LED"));
	comLEDPos = create_combo_box(hbox,_("Position"),led_pos_label,config.led_pos);

	hbox = create_hbox(vboxall);
	txtSnapPath = create_text_with_label(hbox, _("Snapshot Path"), config.snapshot_path, 40);
	create_button(hbox,_("Folder..."), G_CALLBACK(OnSelectSnapPath));

	hbox = create_hbox(vboxall);
	txtFontPath = create_text_with_label(hbox, _("Font Path"), config.font_path, 40);
	create_button(hbox, _("Folder..."), G_CALLBACK(OnSelectFontPath));

	hbox = create_hbox(vboxall);
	txtMsgFontName = create_text_with_label(hbox, _("Message Font"), config.msgboard_msg_fontname, 15);
	sprintf(buf, "%d", config.msgboard_msg_fontsize);
	txtMsgFontSize = create_text_with_label(hbox, _(" Size"), buf, 3);

	hbox = create_hbox(vboxall);
	txtInfoFontName = create_text_with_label(hbox, _("Info Font"), config.msgboard_info_fontname, 15);
	sprintf(buf, "%d", config.msgboard_info_fontsize);
	txtInfoFontSize = create_text_with_label(hbox, _(" Size"), buf, 3);

	hbox = create_hbox(vboxall);
	create_label(hbox, _("* Need restart program."));

	// 2 tape

	vboxall = create_vbox(NULL);
	add_note(nb, vboxall, _("Tape"));

	hboxall = create_hbox(vboxall);

	create_frame(hboxall, _("Load Wav File"), &vbox, &hbox);
	chkReverseWave = create_check_box(hbox, _("Reverse Wave"), config.wav_reverse);
	hbox = create_hbox(vbox);
	chkHalfWave = create_check_box(hbox, _("Half Wave"), config.wav_half);
	hbox = create_hbox(vbox);
	chkCorrectWave = create_check_box(hbox, _("Correct"), config.wav_correct);
	const char *correct_type_label[] = {
		_("COS Wave"),
		_("SIN Wave"),
		NULL
	};
	hbox = create_hbox(vbox);
	create_radio_box(hbox, correct_type_label, 2, radCorrectType, config.wav_correct_type);

	create_frame(hboxall, _("Save Wav File"), &vbox, &hbox);
	const char *rate_label[] = {
		"11025",
		"22050",
		"44100",
		"48000",
		NULL
	};
	comSampleRate = create_combo_box(hbox,_("Sample Rate"),rate_label,config.wav_sample_rate);
	const char *bits_label[] = {
		"8",
		"16",
		NULL
	};
	hbox = create_hbox(vbox);
	comSampleBits = create_combo_box(hbox,_("Sample Bits"),bits_label,config.wav_sample_bits);

	// 3 network

	vboxall = create_vbox(NULL);
	add_note(nb, vboxall, _("Network"));

	// LPT
	for(int drv=0; drv<MAX_PRINTER; drv++) {
		hbox = create_hbox(vboxall);
		sprintf(buf, _("LPT%d Hostname"), drv);
		txtLPTHost[drv] = create_text_with_label(hbox, buf, config.printer_server_host[drv], 20);
		sprintf(buf, "%d", config.printer_server_port[drv]);
		txtLPTPort[drv] = create_text_with_label(hbox, _(" Port"), buf, 5);
	}
	// COM
	const char *comm_baud_label[] = {
		_("S:300baud/F:1200baud"),
		_("S:600baud/F:2400baud"),
		_("S:1200baud/F:4800baud"),
		_("S:2400baud/F:9600baud"),
		NULL
	};
	for(int drv=0; drv<MAX_COMM; drv++) {
		hbox = create_hbox(vboxall);
		sprintf(buf, _("COM%d Hostname"), drv);
		txtCOMHost[drv] = create_text_with_label(hbox, buf, config.comm_server_host[drv], 20);
		sprintf(buf, "%d", config.comm_server_port[drv]);
		txtCOMPort[drv] = create_text_with_label(hbox, _(" Port"), buf, 5);
		comCOMBaud[drv] = create_combo_box(hbox, " ", comm_baud_label, config.comm_dipswitch[drv]);
	}

	// 4 misc

	vboxall = create_vbox(NULL);
	add_note(nb, vboxall, _("Misc"));

	hbox = create_hbox(vboxall);
	txtROMPath = create_text_with_label(hbox, _("ROM Path (*)"), config.rom_path, 40);
	create_button(hbox,_("Folder..."), G_CALLBACK(OnSelectROMPath));

	hbox = create_hbox(vboxall);
	create_label(hbox, _("* Need restart program."));

	//

	gtk_widget_show_all(dialog);
	g_signal_connect(G_OBJECT(dialog), "response", G_CALLBACK(OnResponse), (gpointer)this);
//	gint rc = gtk_dialog_run(GTK_DIALOG(dialog));

	emu->set_pause(1, true);

	return true;
}

void ConfigBox::Hide()
{
	DialogBox::Hide();
	emu->set_pause(1, false);
}

bool ConfigBox::SetData()
{
	int val = 0;

	if (get_check_state(chkMODE)) {
		config.dipswitch |= 4;
	} else {
		config.dipswitch &= ~4;
	}
	emu->set_parami(EMU::ParamFddType, get_radio_state_idx(radFDD, 3));
	emu->set_parami(EMU::ParamIOPort, get_check_state_num(chkIO, 7));

	config.use_opengl = (uint8)get_combo_sel_num(comUseOpenGL);
	config.gl_filter_type = (uint8)get_combo_sel_num(comGLFilter);

	config.disptmg_skew = get_combo_sel_num(comDispSkew);
	config.curdisp_skew = get_combo_sel_num(comCurdSkew);

	config.led_pos = get_combo_sel_num(comLEDPos);

	strcpy(config.snapshot_path, get_text(txtSnapPath));
	UTILITY::add_path_separator(config.snapshot_path);
	strcpy(config.font_path, get_text(txtFontPath));
	UTILITY::add_path_separator(config.rom_path);
	strcpy(config.msgboard_msg_fontname, get_text(txtMsgFontName));
	strcpy(config.msgboard_info_fontname, get_text(txtInfoFontName));
	val = 0;
	if (sscanf(get_text(txtMsgFontSize), "%d", &val) == 1 && val >= 6 && val <= 60) {
		config.msgboard_msg_fontsize = val;
	}
	val = 0;
	if (sscanf(get_text(txtInfoFontSize), "%d", &val) == 1 && val >= 6 && val <= 60) {
		config.msgboard_info_fontsize = val;
	}

	config.wav_reverse = get_check_state(chkReverseWave);
	config.wav_half = get_check_state(chkHalfWave);
	config.wav_correct = get_check_state(chkCorrectWave);
	config.wav_correct_type = (uint8)get_radio_state_idx(radCorrectType, 2);
	config.wav_sample_rate = (uint8)get_combo_sel_num(comSampleRate);
	config.wav_sample_bits = (uint8)get_combo_sel_num(comSampleBits);

	for(int i=0; i<MAX_PRINTER; i++) {
		strcpy(config.printer_server_host[i], get_text(txtLPTHost[i]));
		val = 0;
		if (sscanf(get_text(txtLPTPort[i]), "%d", &val) == 1 && val > 0 && val < 65535) {
			config.printer_server_port[i] = val;
		}
	}
	for(int i=0; i<MAX_COMM; i++) {
		strcpy(config.comm_server_host[i], get_text(txtCOMHost[i]));
		val = 0;
		if (sscanf(get_text(txtCOMPort[i]), "%d", &val) == 1 && val > 0 && val < 65535) {
			config.comm_server_port[i] = val;
		}
		config.comm_dipswitch[i] = get_combo_sel_num(comCOMBaud[i]);
	}

	strcpy(config.rom_path, get_text(txtROMPath));
	UTILITY::add_path_separator(config.rom_path);

	// set message font
	MsgBoard *msgboard = emu->get_msgboard();
	if (msgboard) {
		msgboard->SetFont();
	}

	save_config();
	emu->change_pos_led(config.led_pos);
#ifdef USE_OPENGL
	emu->change_opengl_attr();
#endif

	return true;
}

void ConfigBox::ShowFolderBox(const char *title, GtkWidget *entry)
{
	FolderBox fbox(dialog);
	const char *path = gtk_entry_get_text(GTK_ENTRY(entry));
	if (fbox.Show(title, path)) {
		gtk_entry_set_text(GTK_ENTRY(entry),fbox.GetPath());
	}
}

void ConfigBox::ChangeFDDType(int index)
{
//	int index = get_radio_state_idx(radFDD,3);
//	g_print("ChangeFDDType: %d\n", index);
	if (chkIO[0] == NULL) return;
	if (index == 0) {
		set_check_state(chkIO[0],false);
		set_check_state(chkIO[1],false);
		gtk_widget_set_sensitive(chkIO[0], true);
		gtk_widget_set_sensitive(chkIO[1], true);
	} else if (index == 1) {
		set_check_state(chkIO[0],false);
		set_check_state(chkIO[1],true);
		gtk_widget_set_sensitive(chkIO[0], false);
		gtk_widget_set_sensitive(chkIO[1], false);
	} else if (index == 2) {
		set_check_state(chkIO[0],true);
		set_check_state(chkIO[1],false);
		gtk_widget_set_sensitive(chkIO[0], false);
		gtk_widget_set_sensitive(chkIO[1], false);
	}
}

void ConfigBox::ChangeIOPort(int index)
{
//	g_print("ChangeIOPort: %d\n", index);
	if (chkIO[0] == NULL) return;
	bool status = get_check_state(chkIO[index]);
	if (index == 0 && status) {
		set_check_state(chkIO[index+1],false);
	} else if (index == 1 && status) {
		set_check_state(chkIO[index-1],false);
	}
	if (index == 5 && status) {
		set_check_state(chkIO[index+1],false);
	} else if (index == 6 && status) {
		set_check_state(chkIO[index-1],false);
	}
}

void ConfigBox::OnChangedFDD(GtkWidget *widget, gpointer user_data)
{
	ConfigBox *obj = (ConfigBox *)user_data;
	int index = (int)(intptr_t)g_object_get_data(G_OBJECT(widget), "num");
	obj->ChangeFDDType(index);
}

void ConfigBox::OnChangedIO(GtkWidget *widget, gpointer user_data)
{
	ConfigBox *obj = (ConfigBox *)user_data;
	int index = (int)(intptr_t)g_object_get_data(G_OBJECT(widget), "num");
	obj->ChangeIOPort(index);
}

void ConfigBox::OnSelectSnapPath(GtkWidget *widget, gpointer user_data)
{
	ConfigBox *obj = (ConfigBox *)user_data;
	obj->ShowFolderBox(_("Snapshot Path"), obj->txtSnapPath);
}

void ConfigBox::OnSelectFontPath(GtkWidget *widget, gpointer user_data)
{
	ConfigBox *obj = (ConfigBox *)user_data;
	obj->ShowFolderBox(_("Font Path"), obj->txtFontPath);
}

void ConfigBox::OnSelectROMPath(GtkWidget *widget, gpointer user_data)
{
	ConfigBox *obj = (ConfigBox *)user_data;
	obj->ShowFolderBox(_("Rom Path"), obj->txtROMPath);
}

void ConfigBox::OnResponse(GtkWidget *widget, gint response_id, gpointer user_data)
{
	ConfigBox *obj = (ConfigBox *)user_data;
	if (response_id == GTK_RESPONSE_ACCEPT) {
		obj->SetData();
	}
//	g_print("OnResponse: %d\n",response_id);
	obj->Hide();
}

}; /* namespace GUI_GTK_X11 */


