/** @file msgboard.h

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2012.06.26 -

	@brief [ message board ]
*/

#ifndef _MSGBOARD_H_
#define _MSGBOARD_H_

#include <list>
#include <SDL.h>
#include <SDL_ttf.h>
#include "common.h"

class EMU;

/**
	@brief MsgBoard is the class that display various messages on the screen.
*/
class MsgBoard
{
private:
	EMU *emu;
	SDL_Surface *sMainSuf;

	SDL_Color fg;
	SDL_Color bg;

	struct {
		int cx;
		int cy;
	} szWin;

	typedef struct msg_item_st {
		char cmsg[1024];
		int cnt;
	} item_t;

	typedef std::list<item_t> list_t;

	typedef struct msg_data_st {
		SDL_Rect re;	// `g
		struct {
			int cx;
			int cy;
		} sz;			// `悵̕A
		struct {
			int x;
			int y;
		} pt;			// `̈ʒu
		int place;		// pt̊ 0: 1:E 2: 3:E
		TTF_Font *font;
		list_t lists;
	} msg_data_t;

	msg_data_t msg;
	msg_data_t info;

	bool visible;
	bool enable;

	bool inited;

	typedef struct sbuf_st {
		char buf[1024];
	} sbuf_t;

	// o
	void draw(SDL_Surface *screen, msg_data_t &data);
	// obNobt@ɕ`
	void draw_text(msg_data_t &data);
	void draw_text(SDL_Surface *suf, msg_data_t &data, int left, int top);
	// JEg_E
	void count_down(msg_data_t &data);

	// ʕ\pVXetHg̐ݒ
	bool set_sys_font(const char *title, const _TCHAR *name, int pt, TTF_Font **font);
public:
	MsgBoard(EMU *pEmu);
	~MsgBoard();

	// 
	void CreateSurface(SDL_PixelFormat *format, int width, int height);
	void InitScreen(SDL_PixelFormat *format, int width, int height);

	// \
	void SetVisible(bool val) {
		visible = val;
	}
	// JEg_E
	void CountDown(void);

	// bZ[Wݒ
	void Set(msg_data_t &data, const _TCHAR *str, int sec);
	// bZ[W폜
	void Delete(msg_data_t &data, const _TCHAR *str);
	// EBhETCYݒ
	void SetSize(int width, int height);

	// bZ[Wʒuݒ
	void SetMessagePos(int cx, int cy, int place);
	// bZ[Wݒ
	void SetMessage(const _TCHAR *str, int sec = 5);
	void SetMessageF(const _TCHAR *format, ...);
	// bZ[W폜
	void DeleteMessage(const _TCHAR *str);

	// ʒuݒ
	void SetInfoPos(int cx, int cy, int place);
	// ݒ
	void SetInfo(const _TCHAR *str, int sec = 2);
	void SetInfoF(const _TCHAR *format, ...);
	// 폜
	void DeleteInfo(const _TCHAR *str);

	// o
	void Draw(SDL_Surface *screen);

	// tHgݒ
	bool SetFont();

	// LH
	bool IsEnable() { return enable; }
};

#endif
