/** @file board.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2012.06.21 -

	@brief [ main board ]
*/

#include "board.h"

void BOARD::initialize()
{
}

void BOARD::reset()
{
	now_halt = 0;
	now_nmi  = 0;
	now_irq  = 0;
	now_firq = 0;
	now_wreset = 0;
}

void BOARD::write_signal(int id, uint32 data, uint32 mask)
{
	uint16 prev = 0;
	switch (id) {
		case SIG_CPU_RESET:
			prev = now_wreset;
			now_wreset = ((data & mask) ? (now_wreset | mask) : (now_wreset & ~mask));
			if (prev == 0 && now_wreset != 0) {
				write_signals(&outputs_reset, 0xffffffff);
			} else if (prev != 0 && now_wreset == 0) {
				write_signals(&outputs_reset, 0);
			}
//			emu->out_debugf(_T("BOARD: RESET: %X %X"), data, mask);
			break;
		case SIG_CPU_NMI:
			prev = now_nmi;
			now_nmi = ((data & mask) ? (now_nmi | mask) : (now_nmi & ~mask));
			if (!FUSE_INTR_MASK) {
				if (prev == 0 && now_nmi != 0) {
					write_signals(&outputs_nmi, now_nmi);
				} else if (prev != 0 && now_nmi == 0) {
					write_signals(&outputs_nmi, now_nmi);
				}
//				emu->out_debugf(_T("BOARD: NMI: %X %X"), data, mask);
			}
			break;
		case SIG_CPU_IRQ:
			prev = now_irq;
			now_irq = ((data & mask) ? (now_irq | mask) : (now_irq & ~mask));
			if (!FUSE_INTR_MASK) {
				if (prev == 0 && now_irq != 0) {
					write_signals(&outputs_irq, now_irq);
				} else if (prev != 0 && now_irq == 0) {
					write_signals(&outputs_irq, now_irq);
				}
//				emu->out_debugf(_T("BOARD: IRQ NOW:%X (%X:%X)"), now_irq, data, mask);
			}
			break;
		case SIG_CPU_FIRQ:
			prev = now_firq;
			now_firq = ((data & mask) ? (now_firq | mask) : (now_firq & ~mask));
			if (!FUSE_INTR_MASK) {
				if (prev == 0 && now_firq != 0) {
					write_signals(&outputs_firq, now_firq);
				} else if (prev != 0 && now_firq == 0) {
					write_signals(&outputs_firq, now_firq);
				}
//				emu->out_debugf(_T("BOARD: FIRQ NOW:%X (%X:%X)"), now_firq, data, mask);
			}
			break;
		case SIG_CPU_HALT:
			prev = now_halt;
			now_halt = ((data & mask) ? (now_halt | mask) : (now_halt & ~mask));
			if (prev == 0 && now_halt != 0) {
				write_signals(&outputs_halt, now_halt);
			} else if (prev != 0 && now_halt == 0) {
				write_signals(&outputs_halt, now_halt);
			}
//			emu->out_debugf(_T("BOARD: HALT: %X %X"), data, mask);
			break;
	}
}

uint32 BOARD::read_signal(int id)
{
	return 0xffffffff;
}

// ----------------------------------------------------------------------------

void BOARD::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 0x41;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// save config flags
	memset(&vm_state, 0, sizeof(vm_state));
	vm_state.fdd_type = emu->get_parami(EMU::ParamFddType);
	vm_state.io_port = emu->get_parami(EMU::ParamIOPort);
	vm_state.flags = ((config.sync_irq ? 1 : 0) | (config.realmode_datarec ? 2 : 0));

	vm_state.sys_mode = (uint8)emu->get_parami(EMU::ParamSysMode);
	vm_state.exram_size_num = (uint8)emu->get_parami(EMU::ParamExMemNum);
	vm_state.now_halt = now_halt;
	vm_state.now_nmi  = now_nmi;
	vm_state.now_irq  = now_irq;
	vm_state.now_firq = now_firq;
	vm_state.now_wreset = now_wreset;

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void BOARD::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	// load config flags
	emu->set_parami(EMU::ParamFddType, vm_state.fdd_type);
	emu->set_parami(EMU::ParamIOPort, vm_state.io_port);
	config.sync_irq = ((vm_state.flags & 1) ? true : false);
	config.realmode_datarec = ((vm_state.flags & 2) ? true : false);

	if (vm_state_i.version >= 2) {
		now_halt = vm_state.now_halt;
		now_nmi  = vm_state.now_nmi;
		now_irq  = vm_state.now_irq;
		now_firq = vm_state.now_firq;
		now_wreset = vm_state.now_wreset;
	}
	if (vm_state_ident.version >= 0x41) {
		emu->set_parami(EMU::ParamSysMode, vm_state.sys_mode);
		emu->set_parami(EMU::ParamExMemNum, vm_state.exram_size_num);
	}
}

