/** @file display.cpp

	HITACHI MB-S1 Emulator 'EmuB-S1'
	Skelton for retropc emulator

	@author Sasaji
	@date   2011.06.08 -

	@brief [ display ]
*/

#include "display.h"
#include "../../fileio.h"
#include "../../config.h"

void DISPLAY::initialize()
{
	memset(font, 0, sizeof(font));
	font_rom_loaded = false;
	l3font_rom_loaded = false;
	font_rom_loaded_at_first = false;

	// load rom images
	load_font_rom_file();

	emu->get_rgbformat(&Rmask, &Gmask, &Bmask, &Rshift, &Gshift, &Bshift);

	// create pc palette
	for(int i = 0; i < 2; i++) {
		for(int g = 0; g < 2; g++) {
			for(int r = 0; r < 2; r++) {
				for(int b = 0; b < 2; b++) {
					int bb = b * (i ? 0xff : 0x7f);
					int rr = r * (i ? 0xff : 0x7f);
					int gg = g * (i ? 0xff : 0x7f);
					color16[b][r][g][i] = emu->map_rgbcolor(rr,gg,bb);
				}
			}
		}
	}
	for(int g = 0; g < 4; g++) {
		for(int r = 0; r < 4; r++) {
			for(int b = 0; b < 4; b++) {
				int bb = b * 0x60 - 1;
				int rr = r * 0x60 - 1;
				int gg = g * 0x60 - 1;
				if (bb < 0) bb = 0;
				if (rr < 0) rr = 0;
				if (gg < 0) gg = 0;
				if (b == 3) bb = 0xff;
				if (r == 3) rr = 0xff;
				if (g == 3) gg = 0xff;
				color64[b][r][g] = emu->map_rgbcolor(rr,gg,bb);
//				emu->out_debugf(_T("palette_64[%d][%d][%d]:%x"),b,r,g,color64[b][r][g]);
			}
		}
	}

//	memset(scrntop, 0, sizeof(scrntop));

	screen_left = 0;
	screen_top = 0;
	screen_right = SCREEN_WIDTH + screen_left;
	screen_bottom = SCREEN_HEIGHT + screen_top;
	scrnline1_offset = 0;

	mode_sel = 0;
	ig_mask = true;

	crtc_start_addr = 0x4000;
	cursor_addr = 0;
	cblink = 0;
	disp_cursor = false;
	for(int i=0; i<32; i++) {
		cur_left[i] = 0;
		cur_right[i] = 0;
	}
	cur_top = 0;
	cur_bottom = 0;

	crtc_chr_clocks = 0;
	crtc_ppch = 16;

	interlace = 0;
	videomode = 0;
	disptmg_skew = 0;
	curdisp_skew = 0;

	bg_color = 0;
	hireso = 0;
	width = 1;
	width_sel = 0;
	page_no = 0;
	gwidth = 40;
	gwidth_sel = 0;
	gpage_no = 0;

	top_raster = 0;
	bottom_raster = 0;
	vsync_top_raster = 0;
	vsync_bottom_raster = 0;
	raster_per_char = 16;
	now_raster = 0;
	lines_per_2frames = (LINES_PER_FRAME << 1);

	// synchronizable vertical range
	v_total_min[0] = (LINES_PER_FRAME - 24);
	v_total_max[0] = (LINES_PER_FRAME + 24);
	v_total_min[1] = (LINES_PER_FRAME - 24) / 2;
	v_total_max[1] = (LINES_PER_FRAME + 24) / 2;
	v_total_min[2] = (LINES_PER_FRAME - 24) * 2;
	v_total_max[2] = (LINES_PER_FRAME + 24) * 2;

	for(int i=0; i<VRAM_BUF_HEIGHT; i++) {
		left_col[i] = 0;
		right_col[i] = 0;
		cp_odd_line[i] = 1;
		hsync_left_col[i] = 0;
		hsync_right_col[i] = 0;
	}
	left_diff = 0;
	// synchronizable horizontal range
	h_total_min[0] = CHARS_PER_LINE - 3;	// w.40
	h_total_max[0] = CHARS_PER_LINE + 3;
	h_total_min[1] = CHARS_PER_LINE * 2 - 4;	// w.80
	h_total_max[1] = CHARS_PER_LINE * 2 + 4;

	sw = 0;

	Rloss = (64 << Rshift);
	Gloss = (32 << Gshift);
	Bloss = (96 << Bshift);

	for(int i=0; i<8; i++) {
		ppd[i]=(mode_sel & 0x80) ? 1 : 2;	// pixel per 1dot
		dof[i]=(0x80 >> i);
//		sppd[i]=1;							// pixel per 1dot
//		sdof[i]=(0x80 >> i);
	}
	// pixel per 1char
	ppch = (mode_sel & 0x80) ? 8 : 16;
//	sppch = 8;

	//
	memset(crt_mon, 0, sizeof(crt_mon));
	crt_mon_col = 0;
	crt_mon_row = 0;
	crt_mon_l3vaddr = L3_ADDR_VRAM_START;
	crt_mon_s1vtaddr = 0;
	crt_mon_s1vgaddr = 0;
	crt_mon_raster = 0;
	crt_mon_page = 0;

	crt_mon_cur_col = 0;
	crt_mon_cur_row = 0;

	disp_page_brg_mask.i = 0;

	skip_frame_count = 0;

	// CRTC 2MHz default
	crtc_chr_clocks = 1;
	crtc_ppch = 8;

	update_config();

#if 0
	bg_color = 0x01;

#endif
	// register event (unused)
//	register_frame_event(this);
}

void DISPLAY::reset()
{
	load_font_rom_file();

	REG_INTERLACE_SEL = 0;
	REG_IGENABLE = 0;
	ig_mask = true;

	now_cpu_halt = 0;

	memset(palette_brg, 0, sizeof(palette_brg));

	if (NOW_S1_MODE) {
		width = 40;
		gwidth = 40;	// (320 / 8)
		// dark gray
		color16[0][0][0][1] = emu->map_rgbcolor(0x1f,0x1f,0x1f);
	} else {
		// black 
		color16[0][0][0][1] = emu->map_rgbcolor(0x0,0x0,0x0);
	}
}

void DISPLAY::update_config()
{
	scanline = config.scan_line;
	afterimage = config.afterimage;
//	keepimage = config.keepimage;

	// color filter for afterimage
//	memset(scrnline2, 0x00, sizeof(scrnline2));
	memset(scrnflag,  0x00, sizeof(scrnflag));
	if (afterimage == 0) {
		memset(scrnline1, 0x00, sizeof(scrntype) * SCREEN_WIDTH);
		scrnline1_offset = 0;
	} else if (afterimage == 1) {
		scrnline1_offset = SCREEN_WIDTH;
	} else if (afterimage == 2) {
		scrnline1_offset = SCREEN_WIDTH;
	}

//	if (keepimage > 0) {
//		memset(scrntop, 0, sizeof(scrntop));
//	}
}

void DISPLAY::load_font_rom_file()
{
	const _TCHAR *app_path, *rom_path[2];
	_TCHAR file_path[_MAX_PATH];

	rom_path[0] = config.rom_path;
	rom_path[1] = emu->application_path();

	FILEIO* fio = new FILEIO();
	for(int i=0; i<2; i++) {
		app_path = rom_path[i];

		if (!font_rom_loaded && !l3font_rom_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sFONT.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(font, 0x1000, 1);
				fio->Fclose();
				l3font_rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("FONT.ROM was loaded."));
			}
		}
		if (!font_rom_loaded && !l3font_rom_loaded) {
			UTILITY::stprintf(file_path, _MAX_PATH, _T("%sS1FONT.ROM"), app_path);
			if(fio->Fopen(file_path, FILEIO_READ_BINARY)) {
				fio->Fread(font, 0x2000, 1);
				fio->Fclose();
				font_rom_loaded = true;
				emu->out_log(LOG_INFO, (CT)_("S1FONT.ROM was loaded."));
			}
		}
		if (font_rom_loaded || l3font_rom_loaded) break;
	}
	delete fio;

	if (!l3font_rom_loaded) {
		if (!font_rom_loaded_at_first) {
			emu->out_log(LOG_WARN, (CT)_("FONT.ROM couldn't be loaded."));
		}
		if (!font_rom_loaded) {
			if (!font_rom_loaded_at_first) {
				emu->out_log(LOG_WARN, (CT)_("S1FONT.ROM couldn't be loaded."));
			}
			memset(chrfont, 0, sizeof(chrfont));
		} else {
			// convert font data to useful array data
			set_font_data();
			conv_to_l3font_data();
			font_rom_loaded = false;
			l3font_rom_loaded = true;
		}
	} else {
		// convert font data to useful array data
		set_l3font_data();
	}
	font_rom_loaded_at_first = true;
}

void DISPLAY::write_io8(uint32 addr, uint32 data)
{
	addr &= 0xffff;
	switch(addr) {
	case 0xfe20:
		// bmsk color graphic color
		REG_BMSK_COLOR = (data & 0x87);
		SIG_BMSK_DW = (BMSK_COLOR_BMSK ? (SIG_BMSK_DW | 2) : (SIG_BMSK_DW & ~2));
		break;
	case 0xfe21:
		// active graphic plane
		REG_ACTIVE_PLANE = (data & 0x0f);
		SIG_BMSK_DW = (ACTIVE_PLANE_DW ? (SIG_BMSK_DW | 1) : (SIG_BMSK_DW & ~1));
		break;
	case 0xfe23:
		// display page number
		REG_DISP_PAGE = (data & 0xff);

		set_disp_page();
		break;
	case 0xfe24:
		// scrn mode
		REG_SCRN_MODE = (data & 0xff);

		if (NOW_S1_MODE) {
			set_s1_screen_mode();
		}
		break;
	case 0xfe25:
	case 0xfe26:
	case 0xfe27:
		// graphic color
		REG_GCOLOR[addr - 0xfe25] = (data & 0xff);
		break;
	case 0xfe28:
	case 0xfe29:
	case 0xfe2a:
	case 0xfe2b:
	case 0xfe2c:
	case 0xfe2d:
	case 0xfe2e:
	case 0xfe2f:
		// palette
		REG_PALETTE[addr - 0xfe28] = (data & 0x0f);
		set_palette_brg(addr - 0xfe28);

//		emu->out_debugf("dispw a:%04x d:%02x",addr,data);
		break;
	case 0xffc6:
		// crtc addr register
		crtc_addr = (data & 0xff);
		break;
	case 0xffc7:
		// crtc data register
		switch(crtc_addr) {
			case 1:
				if (NOW_L3_MODE) {
					width = regs[1] ? regs[1] : 1;				// width
				}
				break;
			case 8:
				interlace = (regs[8] & 1) ? 1 : 0;				// 0:noninterlace 1:interlace
				videomode = ((regs[8] & 3) == 3) ? 1 : 0;		// 1:videomode(interlace only)

				disptmg_skew = ((regs[8] & 0x30) >> 4);			// disptmg skew
				curdisp_skew = ((regs[8] & 0xc0) >> 6);			// curdisp skew
//				break;	// through
			case 9:
				// raster of 1char
				raster_per_char = (regs[9] & 0x1f) + videomode + 1;
				raster_per_char = raster_per_char << (1 - videomode);
				break;

		}
		break;
	case 0xffd0:
		// mode sel
		mode_sel = (data & 0xff);

		bg_color = mode_sel & 0x0f;				// bg color

		if (NOW_L3_MODE) {
			bg_color |= 0x08;					// bright

			hireso = (mode_sel & 0x40) ? 0 : 1;		// 0:normal 1:hireso
			width_sel = (mode_sel & 0x80) ? 1 : 0;	// 0:w.40  1:w.80

			for(int i=0; i<8; i++) {
				ppd[i]=width_sel ? 1 : 2;	// pixel per 1dot
				dof[i]=(0x80 >> i);
			}
			// pixel per 1char
			ppch = (mode_sel & 0x80) ? 8 : 16;
		}
		break;
	case 0xffd6:
		// interlace sel
		REG_INTERLACE_SEL = ((data & 0x08) >> 3);	// 0:noninterlace char  1:interlace char
		break;
	case 0xffe9:
		// ig mode reg
		REG_IGENABLE = (data & 0x01);
		ig_mask &= ((data & 0x100) == 0);
//		REG_IGMODE = (ig_mask ? (REG_IGMODE & ~0x80) : (REG_IGMODE | 0x80));
		break;
	}
}

uint32 DISPLAY::read_io8(uint32 addr)
{
	uint32 data = 0xff;
	switch(addr & 0xffff) {
	case 0xfe28:
	case 0xfe29:
	case 0xfe2a:
	case 0xfe2b:
	case 0xfe2c:
	case 0xfe2d:
	case 0xfe2e:
	case 0xfe2f:
		// palette cannot be read
		// when disptmg is on, send HALT signal 
//		if (DISPTMG_ON && !now_cpu_halt) {
		// when hsync and vsync is off, send HALT signal 
		if (!HVSYNC_ON && !now_cpu_halt) {
			d_board->write_signal(SIG_CPU_HALT, SIG_HALT_PALETTE_MASK, SIG_HALT_PALETTE_MASK);
			now_cpu_halt = 1;
//			emu->out_debugf(_T("palette halt fired: %d"),data);
		}
		break;
	}
	return data;
}

void DISPLAY::write_signal(int id, uint32 data, uint32 mask)
{
	switch(id) {
		// Lightpen not supported
#if 0
		case SIG_DISPLAY_LPSTB:
			{
			int width = regs[1];
			int dotx = (data >> 16) & 0xffff;
			int doty = (data & 0xffff);
			int x = 0;
			int y = 0;
			int offset = (width_sel ? 5 : 4);
			uint32 newaddr;

			if (dotx < left_col[doty >> 1]) { dotx = right_col[doty >> 1] - 1; doty-=16; }
			if (dotx >= right_col[doty >> 1]) { dotx = left_col[doty >> 1]; doty+=16; }
			if (doty < top_raster) doty = top_raster;
			if (doty >= bottom_raster) doty = bottom_raster - 1;

			x = (dotx - left_col[doty >> 1]) >> (width_sel ? 3 : 4);
			y = (doty - top_raster) >> 4;
			newaddr = ((regs[12] << 8)+regs[13]) + y * width + x + offset;
			regs[16] = (newaddr >> 8) & 0x3f;
			regs[17] = newaddr & 0xff;

//			emu->out_debugf("lp dx:%03d dy:%03d x:%02d y:%02d addr:%02x%02x"
//				,dotx,doty,x,y,regs[16],regs[17]);
			}
			break;
#endif
		case SIG_DISPLAY_DISPTMG:
			// display area
			REG_IGMODE = ((data & mask) ? (REG_IGMODE | DISPTMG_MASK) : (REG_IGMODE & ~DISPTMG_MASK));
//			emu->out_debugf(_T("disptmg %06lld IGM:%02x  %d:%d"), current_clock() & 0xffffff, REG_IGMODE, data, mask);
			break;
		case SIG_DISPLAY_HSYNC:
			// hsync
			REG_IGMODE = ((data & mask) ? (REG_IGMODE | HSYNC_MASK) : (REG_IGMODE & ~HSYNC_MASK));
			// release halt signal
			if (now_cpu_halt) {
				now_cpu_halt = 0;
				d_board->write_signal(SIG_CPU_HALT, 0, SIG_HALT_PALETTE_MASK);
//				emu->out_debugf(_T("hsync halt released: %d %d"),data,mask);
			}
			// TODO: PAE flag is different from real machine
			REG_IGMODE = (HVSYNC_ON ? REG_IGMODE | PAE_MASK : REG_IGMODE & ~PAE_MASK);

//			emu->out_debugf(_T("hsync   %06lld IGM:%02x  %d:%d"), current_clock() & 0xffffff, REG_IGMODE, data, mask);
			break;
		case SIG_DISPLAY_VSYNC:
			if ((data & mask) == 0) {
				REG_IGMODE = (REG_IGMODE & ~VSYNC_MASK);
				// synchronize display vertical raster
				for(int i=0; i<3; i++) {
					if ((v_total_min[i] <= now_raster && now_raster <= v_total_max[i])
					|| (v_total_min[i] <= *crtc_vt_total && *crtc_vt_total <= v_total_max[i])) {
						now_raster = 0;
						break;
					}
				}
				// cursor blink
				cblink = (cblink + 1) & 0x1f;

			} else {
				REG_IGMODE = (REG_IGMODE | VSYNC_MASK);
			}
			// release halt signal
			if (now_cpu_halt) {
				now_cpu_halt = 0;
				d_board->write_signal(SIG_CPU_HALT, 0, SIG_HALT_PALETTE_MASK);
//				emu->out_debugf(_T("vsync halt released: %d %d"),data,mask);
			}
			// TODO: PAE flag is different from real machine
			REG_IGMODE = (HVSYNC_ON ? REG_IGMODE | PAE_MASK : REG_IGMODE & ~PAE_MASK);

//			emu->out_debugf(_T("vsync   %06lld IGM:%02x  %d:%d ras:%d"), current_clock() & 0xffffff, REG_IGMODE, data, mask, now_raster);
			break;
		case SIG_DISPLAY_CHR_CLOCKS:
			crtc_chr_clocks = (data == CLOCKS_1MHZ ? 0 : 1);
			crtc_ppch = (crtc_chr_clocks ? 8 : 16);
			break;
		case SIG_CPU_RESET:
			// halt on while res signal is on.
			if (!now_reset && (data & mask) && now_cpu_halt == 0) {
				// turn on
				d_board->write_signal(SIG_CPU_HALT, SIG_HALT_PALETTE_MASK, SIG_HALT_PALETTE_MASK);
				now_cpu_halt = 1;
			} else if (now_reset && (data & mask) == 0 && now_cpu_halt) {
				// turn off
				d_board->write_signal(SIG_CPU_HALT, 0, SIG_HALT_PALETTE_MASK);
				now_cpu_halt = 0;
			}
			now_reset = (data & mask) ? true : false;
			if (!now_reset) {
				if (NOW_S1_MODE) {
					// dark gray
					color16[0][0][0][1] = emu->map_rgbcolor(0x1f,0x1f,0x1f);
				} else {
					// black 
					color16[0][0][0][1] = emu->map_rgbcolor(0x0,0x0,0x0);
				}
			}
			break;
	}
}

void DISPLAY::update_cursor()
{
	int bp = regs[10] & 0x60;						// cursor blink speed
	int cur_left_0;
	int cur_width;
	uint8 cur_st_l;
	uint8 cur_ed_l;

	if(crt_mon_cur_col < 0 || crt_mon_cur_row < 0 || curdisp_skew == 3) {
		disp_cursor = false;
	} else {
		disp_cursor = (bp == 0 || (bp == 0x40 && (cblink & 8)) || (bp == 0x60 && (cblink & 0x10)));
	}

	if (disp_cursor) {
		if (videomode == 1) {
			cur_st_l = (regs[10] & 0x0f);
			cur_ed_l = (regs[11] & 0x0f);
		} else {
			cur_st_l = ((regs[10] & 0x0f) << 1);
			cur_ed_l = ((regs[11] & 0x0f) << 1) + 1;
		}

		// cursor draw area
		cur_top = top_raster + crt_mon_cur_row;
		cur_bottom = cur_top + cur_ed_l;
		cur_top += cur_st_l;

		cur_left_0 = left_col[crt_mon_cur_row] + (crt_mon_cur_col + curdisp_skew + (NOW_L3_MODE ? config.curdisp_skew : 0) - 1 - disptmg_skew + config.disptmg_skew) * crtc_ppch;
		cur_width = crtc_ppch + ((NOW_S1_MODE && width_sel == 0) ? crtc_ppch : 0);
		for(int i=0; i < raster_per_char; i++) {
			cur_left[i] = cur_left_0;
			cur_right[i] = cur_left[i] + cur_width;
		}
	}
	crt_mon_cur_col = -1;
	crt_mon_cur_row = -1;
}

void DISPLAY::set_display_size(int left, int top, int right, int bottom)
{
	screen_left = left;
	screen_right = right;
	screen_top = top;
	screen_bottom = bottom;
}

void DISPLAY::draw_screen()
{
	if (emu->now_skip_frame()) {
		skip_frame_count++;
	} else {
		skip_frame_count = 0;
	}
	if ((afterimage == 0 && skip_frame_count >= 1)
	||  (afterimage > 0 && skip_frame_count >= 2)) {
		// clear cursor
		disp_cursor = false;
		crt_mon_cur_col = -1;
		crt_mon_cur_row = -1;

		return;
	}

	sw = 1 - sw;

	// update cursor
	update_cursor();

	// render screen
	draw_screen_sub();
}

uint8 DISPLAY::get_font_data(int interlace, int chr_code, int y)
{
	int code = 0;
	int offset = 0;
	if (interlace == 1) {
		// interlace mode char
		code = (chr_code | 0x100) << 4;	// 16byte per 1char
		offset = (y & 0xf);	// 0 - 15
	} else {
		// non interlace mode char
		code = chr_code << 4;	// 16byte per 1char
		offset = (y & 0x7);	// 0 - 7
	}
	return font[code + offset];
}

void DISPLAY::set_font_data()
{
	for(int i = 0; i < 2; i++) {
		for(int c = 0; c < 256; c++) {
			for(int l = 0; l < 16; l++) {
				chrfont[i][c][l] = get_font_data(i, c, l);
			}
		}
	}
}

uint8 DISPLAY::get_l3font_data(int interlace, int chr_code, int y)
{
	int code = 0;
	int offset = 0;
	if (interlace == 1) {
		// interlace mode char
		if (chr_code < 0x80) {
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0e);	// 0 0 2 2 4 4 6 6 8 8 a a c c e e
		} else {
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0f);	// 0 1 2 3 4 5 6 7 8 9 a b c d e f
		}
	} else {
		// non interlace mode char
		if (chr_code < 0x80) {
			// ascii code
			code = chr_code << 4;	// 16byte per 1char
		} else {
			code = ((chr_code - 0x80) << 4) + 1;
		}
		offset = ((y << 1) & 0x0e);	// 0 2 4 6 8 a c e 0 2 4 6 8 a c e
	}
	return font[code + offset];
}

void DISPLAY::set_l3font_data()
{
	for(int i = 0; i < 2; i++) {
		for(int c = 0; c < 256; c++) {
			for(int l = 0; l < 16; l++) {
				chrfont[i][c][l] = get_l3font_data(i, c, l);
			}
		}
	}
}

void DISPLAY::conv_to_l3font_data()
{
	int code = 0;
	int offset = 0;
	memset(font, 0, sizeof(font));
	for(int chr_code = 0; chr_code < 256; chr_code++) {
		for(int y=0; y<8; y++) {
			// non interlace mode char
			if (chr_code < 0x80) {
				// ascii code
				code = chr_code << 4;	// 16byte per 1char
			} else {
				code = ((chr_code - 0x80) << 4) + 1;
			}
			offset = ((y << 1) & 0x0e);	// 0 2 4 6 8 a c e 0 2 4 6 8 a c e
			font[code + offset] = chrfont[0][chr_code][y];
		}
	}
	for(int chr_code = 128; chr_code < 256; chr_code++) {
		for(int y=0; y<16; y++) {
			// interlace mode char
			code = chr_code << 4;	// 16byte per 1char
			offset = (y & 0x0f);	// 0 1 2 3 4 5 6 7 8 9 a b c d e f
			font[code + offset] = chrfont[1][chr_code][y];
		}
	}
}

#define COLOR_REVERSE(fg,bg) { \
	uint8 c; \
	c = (bg); \
	(bg) = (fg); \
	(fg) = c; \
}

#define DECODE_TO_PALETTE16_NUM(code, outc) { \
	outc.b.b = (code & 1) ? 1 : 0; \
	outc.b.r = (code & 2) ? 1 : 0; \
	outc.b.g = (code & 4) ? 1 : 0; \
	outc.b.i = (code & 8) ? 1 : 0; \
}

#define DECODE_TO_PALETTE64_NUM(code, outc) { \
	outc.b.b = (code & 1) ? ((code & 8) ? 3 : 2) : 0; \
	outc.b.r = (code & 2) ? ((code & 8) ? 3 : 2) : 0; \
	outc.b.g = (code & 4) ? ((code & 8) ? 3 : 2) : 0; \
}

#define DISP_OUT_OF_DISPTMG { \
	if (hsync_left_col[buf_y_h] <= x && x < hsync_right_col[buf_y_h]) { \
		p_p.i = 0; \
	} else { \
		p_p = b_p; \
	} \
	scrn[x] = (color16[p_p.b.b][p_p.b.r][p_p.b.g][p_p.b.i] | (scrnline1p[x] & scrnline2m)); \
	scrnf[x] &= scrnline2m; \
}

#define DISP_TEXT_CHAR_ONE_DOT(pos, code) { \
	p_p = (code & dof[pos]) ? t_p : b_p; \
	dot = color16[p_p.b.b][p_p.b.r][p_p.b.g][p_p.b.i]; \
	scrnline2m = ((p_p.i & BRG_COLOR_IMASK) ? 0 : 0xffffffff); \
	dtm = x+bi+ppd[pos]; \
	if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
	for(dt = x+bi; dt < dtm; dt++) { \
		scrn[dt] = (dot | (scrnline1p[dt] & scrnline2m)); \
		scrnf[dt] &= scrnline2m; \
	} \
}

#define DISP_TEXT_CHAR_MAIN(code) { \
	bi = 0; \
	for(bt = 0; bt < 8; bt++) { \
		DISP_TEXT_CHAR_ONE_DOT(bt, code); \
		bi += ppd[bt]; \
	} \
}

#define DISP_L3_TEXT_CHAR(code) { \
	pattern = chrfont[REG_INTERLACE_SEL][code][(l >> (1 - videomode)) & 0xf]; \
\
	if (disp_cursor_char) pattern = ~pattern; \
	if (text_cram & 8) pattern = ~pattern; \
	DISP_TEXT_CHAR_MAIN(pattern) \
}

#define DISP_IG_PATTERN_ONE_DOT(ci, bt) { \
	if (disp_cursor_char) { \
		p_p = t_p; \
	} else { \
		pattern = ((ci[0].b.g & dof[bt]) ? 4 : 0) | ((ci[0].b.r & dof[bt]) ? 2 : 0) | ((ci[0].b.b & dof[bt]) ? 1 : 0); \
		DECODE_TO_PALETTE16_NUM(pattern, p_p); \
		if (p_p.i == 0) p_p = b_p; \
		p_p.b.i = 1; \
	} \
	dot = color16[p_p.b.b][p_p.b.r][p_p.b.g][p_p.b.i]; \
	scrnline2m = ((p_p.i & BRG_COLOR_IMASK) ? 0 : 0xffffffff); \
	dtm = x+bi+ppd[bt]; \
	if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
	for(dt = x+bi; dt < dtm; dt++) { \
		scrn[dt] = (dot | (scrnline1p[dt] & scrnline2m)); \
		scrnf[dt] &= scrnline2m; \
	} \
}

#define DISP_IG_PATTERN_MAIN(ci) { \
	bi = 0; \
	for(bt = 0; bt < 8; bt++) { \
		DISP_IG_PATTERN_ONE_DOT(ci, bt) \
		bi += ppd[bt]; \
	} \
}

#define DISP_L3_IG_PATTERN(code) { \
	ll = l >> int_vid; \
	ci[0].b.b = ig_ram[code*8+ll]; \
	ci[0].b.r = ig_ram[code*8+ll+IGRAM_SIZE]; \
	ci[0].b.g = ig_ram[code*8+ll+IGRAM_SIZE+IGRAM_SIZE]; \
	if (text_cram & 8) ci[0].i = ~ci[0].i; \
\
	DISP_IG_PATTERN_MAIN(ci) \
}

#define DISP_WHITE_SPACE { \
	c_f = 15; \
	DECODE_TO_PALETTE16_NUM(c_f, p_p); \
	dot = color16[p_p.b.b][p_p.b.r][p_p.b.g][p_p.b.i]; \
	scrnline2m = ((p_p.i & BRG_COLOR_IMASK) ? 0 : 0xffffffff); \
	dtm = x+ppch; \
	if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
	for(dt = x; dt < dtm; dt++) { \
		scrn[dt] = (dot | (scrnline1p[dt] & scrnline2m)); \
		scrnf[dt] &= scrnline2m; \
	} \
}

#define	DISP_L3_GRAPHIC_HIRESO_MAIN(code) { \
	bi = 0; \
	for(bt = 0; bt < 8; bt++) { \
		p_p = (code & dof[bt]) ? t_p : b_p; \
		dot = color16[p_p.b.b][p_p.b.r][p_p.b.g][p_p.b.i]; \
		scrnline2m = ((p_p.i & BRG_COLOR_IMASK) ? 0 : 0xffffffff); \
		dtm = x+bi+ppd[bt]; \
		if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
		for(dt = x+bi; dt < dtm; dt++) { \
			scrn[dt] = (dot | (scrnline1p[dt] & scrnline2m)); \
			scrnf[dt] &= scrnline2m; \
		} \
		bi += ppd[bt]; \
	} \
}

#define	DISP_L3_GRAPHIC_HIRESO(code) \
	DISP_L3_GRAPHIC_HIRESO_MAIN(code)


#define DISP_L3_GRAPHIC_NORMAL_MAIN(code) { \
	bi = 0; \
	for (bt = 0; bt < 8; bt += 4) { \
		p_p = (code & ((0x10 >> bt) << (l >> 2))) ? t_p : b_p; \
		dot = color16[p_p.b.b][p_p.b.r][p_p.b.g][p_p.b.i]; \
		scrnline2m = ((p_p.i & BRG_COLOR_IMASK) ? 0 : 0xffffffff); \
		dtm = x+bi+ppd[bt]+ppd[bt+1]+ppd[bt+2]+ppd[bt+3]; \
		if (dtm > SCREEN_WIDTH) dtm = SCREEN_WIDTH; \
		for(dt = x+bi; dt < dtm; dt++) { \
			scrn[dt] = (dot | (scrnline1p[dt] & scrnline2m)); \
			scrnf[dt] &= scrnline2m; \
		} \
		bi += ppd[bt] + ppd[bt + 1] + ppd[bt + 2] + ppd[bt + 3]; \
	} \
}

#define DISP_L3_GRAPHIC_NORMAL(code) \
	DISP_L3_GRAPHIC_NORMAL_MAIN(code)

#define EXPAND_BRG_1BIT(code, in8, outc) { \
	if (code & 1) { outc.b.b |= in8; } \
	if (code & 2) { outc.b.r |= in8; } \
	if (code & 4) { outc.b.g |= in8; } \
	if (code & 8) { outc.b.i |= in8; } \
}

#define EXPAND_BRG_2BIT(code, in8, outc) { \
	if (code & 1) { outc[0].b.b |= in8; } \
	if (code & 2) { outc[0].b.r |= in8; } \
	if (code & 4) { outc[0].b.g |= in8; } \
	if (code & 8) { outc[0].b.i |= in8; } \
	outc[1].i = outc[0].i; \
}

#define EXPAND_BRG_2BIT_I(code, in8, outc) { \
	wk1 = in8; wk2 = (code & 8); \
	if (code & 1) { outc[0].b.b |= wk1; outc[1].b.b |= (wk2 ? wk1 : 0); } \
	if (code & 2) { outc[0].b.r |= wk1; outc[1].b.r |= (wk2 ? wk1 : 0); } \
	if (code & 4) { outc[0].b.g |= wk1; outc[1].b.g |= (wk2 ? wk1 : 0); } \
}

#define INIT_BRG_1(data,out) \
	out.i = data;

#define INIT_BRG_2(data,out) \
	out[0].i = out[1].i = data;

#define ENCODE_PALETTE_ONE_DOT_1(pos, inc, outc) { \
	iwk1.i = ((inc.i & (0x01010101 << pos)) >> pos); \
	iwk1.i = palette_brg[iwk1.b.b][iwk1.b.r][iwk1.b.g].i; \
	outc.i |= (iwk1.i << pos); \
}

#define ENCODE_PALETTE_ONE_DOT_2(pos, inc, outc) { \
	iwk1.i = ((inc[0].i & (0x01010101 << pos)) >> pos); \
	iwk1.i = palette_brg[iwk1.b.b][iwk1.b.r][iwk1.b.g].i; \
	outc[0].i |= (iwk1.i << pos); \
	iwk1.i = ((inc[1].i & (0x01010101 << pos)) >> pos); \
	iwk1.i = palette_brg[iwk1.b.b][iwk1.b.r][iwk1.b.g].i; \
	outc[1].i |= (iwk1.i << pos); \
}

#define ENCODE_PALETTE_ONE_DOT_I(pos, inc, outc) { \
	iwk1.i = (inc.i & (0x01010101 << pos)); \
	iwk1.i >>= pos; \
	outc.i = palette_brg[iwk1.b.b][iwk1.b.r][iwk1.b.g].i; \
}
#define DECODE_BRG_2BIT_I(pos, inc, outc) { \
	iwk2.i = (inc.i << pos); \
	if (inc.i == 0x1000000) { \
		outc[1].i |= (0x010101 << pos); \
	} else { \
		outc[0].i |= iwk2.i; \
		if (inc.b.i) outc[1].i |= iwk2.i; \
	} \
}

#define CREATE_MASK_1(inc,outc) { \
	wk1 = ~(inc.b.b | inc.b.r | inc.b.g | inc.b.i); \
	outc.b.b = outc.b.r = outc.b.g = outc.b.i = wk1; \
}

#define CREATE_MASK_2(inc,outc) { \
	wk1 = ~(inc[0].b.b | inc[0].b.r | inc[0].b.g | inc[0].b.i | inc[1].b.b | inc[1].b.r | inc[1].b.g | inc[1].b.i); \
	outc.b.r = outc.b.g = outc.b.b = outc.b.i = wk1; \
}

#define COPY_BRG_1(inc,outc) { \
	outc.i = inc.i; \
}

#define COPY_BRG_2(inc,outc) { \
	outc[0].i = inc[0].i; outc[1].i = inc[1].i; \
}

#define COPY_BRG_1_WITH_MASK(inc,outc,msk) { \
	outc.i = (inc.i & msk.i); \
}

#define COPY_BRG_2_WITH_MASK(inc,outc,msk) { \
	outc[0].i = (inc[0].i & msk.i); outc[1].i = (inc[1].i & msk.i); \
}

#define ADD_BRG_1(inc,outc) { \
	outc.i |= inc.i; \
}

#define ADD_BRG_2(inc,outc) { \
	outc[0].i |= inc[0].i; outc[1].i |= inc[1].i; \
}

#define ADD_BRG_1_WITH_MASK(inc,outc,msk) { \
	outc.i |= (inc.i & msk.i); \
}

#define ADD_BRG_2_WITH_MASK(inc,outc,msk) { \
	outc[0].i |= (inc[0].i & msk.i); outc[1].i |= (inc[1].i & msk.i); \
}

#define REVERSE_BRG_1(outc) { \
	outc.i = ~outc.i; \
}

#define REVERSE_BRG_2(outc) { \
	outc[0].i = ~outc[0].i;	outc[1].i = ~outc[1].i; \
}

#define DECODE_BRG_1BIT_TO_PALETTE16(pos,inc,outc) { \
	iwk1.i = (0x80808080 >> pos); iwk2.i = (7 - pos); \
	outc.i = ((inc[0].i & iwk1.i) >> iwk2.i); \
}

#define DECODE_BRG_2BIT_TO_PALETTE64_R(pos,inc,outc) { \
	iwk1.i = (0x80808080 >> pos); iwk2.i = (7 - pos); \
	outc.i = (((inc[0].i & iwk1.i) >> iwk2.i) | (((inc[1].i & iwk1.i) >> iwk2.i) << 1)); \
}

#define DECODE_BRG_2BIT_TO_PALETTE64(pos,inc,outc) { \
	iwk1.i = (0x80808080 >> pos); iwk2.i = (7 - pos); \
	outc.i = ((((inc[0].i & iwk1.i) >> iwk2.i) << 1) | ((inc[1].i & iwk1.i) >> iwk2.i)); \
}

#define EXPAND_UINT8_LEFT(inp,outp) { \
	outp = ((inp) & 0x01) | (((inp) & 0x02) << 1) | (((inp) & 0x04) << 2) | (((inp) & 0x08) << 3); \
	outp |= ((outp) << 1); \
}							
#define EXPAND_UINT8_RIGHT(inp,outp) { \
	outp = (((inp) & 0x10) >> 4) | (((inp) & 0x20) >> 3) | (((inp) & 0x40) >> 2) | (((inp) & 0x80) >> 1); \
	outp |= ((outp) << 1); \
}

#define EXPAND_UINT32_LEFT(inp,outp) { \
	outp = ((inp) & 0x01010101) | (((inp) & 0x02020202) << 1) | (((inp) & 0x04040404) << 2) | (((inp) & 0x08080808) << 3); \
	outp |= ((outp) << 1); \
}							
#define EXPAND_UINT32_RIGHT(inp,outp) { \
	outp = (((inp) & 0x10101010) >> 4) | (((inp) & 0x20202020) >> 3) | (((inp) & 0x40404040) >> 2) | (((inp) & 0x80808080) >> 1); \
	outp |= ((outp) << 1); \
}

#define SCRNTYPE_SELDOT(newdot, olddot) \
	(afterimage == 0 ? newdot : \
	((newdot & Rmask) > (olddot & Rmask) ? (newdot & Rmask) : (olddot & Rmask)) \
	| ((newdot & Gmask) > (olddot & Gmask) ? (newdot & Gmask) : (olddot & Gmask)) \
	| ((newdot & Bmask) > (olddot & Bmask) ? (newdot & Bmask) : (olddot & Bmask)))


void DISPLAY::draw_screen_sub()
{
	static uint8 pattern;
	static uint8 text_code;
	static uint8 c_f;
	static uint8 c_b;
	static uint8 text_cram;
	static uint8 attr;
	static uint8 fg;
	static uint8 wk1;
	static t_brg_color iwk1,iwk2;
	static t_brg_color ci[2];
//	static uint8 i_r[2],i_g[2],i_b[2];
	static t_brg_color ct[2],ct_m;	// text pattern (r,g,b,mask)
	static t_brg_color cg[2],cg_m;	// grachic pattern (r,g,b,mask)
	static t_brg_color cd[2],cd_m;	// draw pattern (r,g,b,mask)
	static t_brg_color cd2[2];		// draw pattern (r,g,b,mask)
	static t_brg_color cb;			// bg pattern (r,g,b,mask)
	static t_brg_color b_p;			// bg palette num (r,g,b)
	static t_brg_color t_p;			// text palette num (r,g,b)
	static t_brg_color p_p;			// palette num (r,g,b)
	static int bt;
	static int bi;
	static int dt, dtm;
	static int x,x0,x1,xodd;
	static int y,y0[3],y1[3],ys,yd;
	static int l,ll;
	static int int_vid;
	static bool disp_cursor_line;
	static bool disp_cursor_char;
//	static bool ki;
//	bool skip_line;
	static int buf_x = 0;
	static int buf_y;
	static int buf_y_h,buf_y_h2;
	static int i,ii[2];

	static int scrn_offset;
	static scrntype *scrn0;
	static scrntype *scrn;
	static uint8 *scrnf;
	static scrntype dot;
	static uint32 r,g,b;

	static scrntype *scrnline1p;
	static scrntype scrnline2m;

	disp_cursor_line = false;
	disp_cursor_char = false;
	scrn_offset = emu->screen_buffer_offset();
	scrn0 = emu->screen_buffer(0);
	dot = 0;
	buf_x = 0;

	// interace or videomode
	int_vid = ((REG_INTERLACE_SEL) + (1 - videomode));

	// bg color
	INIT_BRG_1(0,cb);
	EXPAND_BRG_1BIT(bg_color, 0xff, cb);
	DECODE_TO_PALETTE16_NUM(bg_color, b_p);

	// color filter for afterimage
	scrnline2m = 0;
	if (afterimage == 1) {
		scrn = scrn0 + scrn_offset * screen_top;
		for(y = screen_top; y < screen_bottom; y++) {
			for(x = screen_left; x < screen_right; x++) {
				dot =scrn[x];
				r = (dot & Rmask);
				g = (dot & Gmask);
				b = (dot & Bmask);

				if (r > Rloss) r = r - Rloss;
				else r = 0;
				if (g > Gloss) g = g - Gloss;
				else g = 0;
				if (b > Bloss) b = b - Bloss;
				else b = 0;

				dot = r | g | b;

				scrnline1[y][x] = dot;
			}
			scrn += scrn_offset;
		}
	} else if (afterimage == 2) {
		scrn = scrn0 + scrn_offset * screen_top;
		for(y = screen_top; y < screen_bottom; y++) {
			for(x = screen_left; x < screen_right; x++) {
				dot = scrn[x];

				if (scrnflag[y][x] > 2) {
					scrnflag[y][x] = 0;
					dot = 0;
				} else {
					scrnflag[y][x]++;
				}
				
				scrnline1[y][x] = dot;
			}
			scrn += scrn_offset;
		}
	}

	// scanline
	ys = (scanline ? 2 : 1);	// skip step
	yd = (scanline && (interlace == 1 && sw == (screen_top & 1)) ? 1 : 0);	// odd line ?

	// set y axis position from the kind of drawing line
	if (disptmg_skew == 3) {
		y0[0] = (screen_top | yd); y1[0] = screen_bottom;	// no drawing
		y0[1] = 0; y1[1] = 0;
		y0[2] = 0; y1[2] = 0;
		ii[0] = 1; ii[1] = 1;
	}
	else if (top_raster <  bottom_raster) {
		y0[0] = (screen_top | yd); y1[0] = (screen_bottom < top_raster ? screen_bottom : top_raster);	// top padding
		y0[1] = (bottom_raster | yd); y1[1] = screen_bottom;	// bottom padding
		y0[2] = ((top_raster > screen_top ? top_raster : screen_top) | yd); y1[2] = (screen_bottom < bottom_raster ? screen_bottom : bottom_raster);
		ii[0] = 2; ii[1] = 3;
	}
	else if (top_raster >= bottom_raster) {
		y0[0] = ((bottom_raster > screen_top ? bottom_raster : screen_top) | yd); y1[0] = (screen_bottom < top_raster ? screen_bottom : top_raster);
		y0[1] = (screen_top | yd); y1[1] = (screen_bottom < bottom_raster ? screen_bottom : bottom_raster);	// top drawing
		y0[2] = (top_raster | yd); y1[2] = screen_bottom;	// bottom drawing
		ii[0] = 1; ii[1] = 3;
	}
	else {
		y0[0] = (screen_top | yd); y1[1] = screen_bottom;
		y0[1] = 0; y1[1] = 0;
		y0[2] = 0; y1[2] = 0;
		ii[0] = 0; ii[1] = 1;
	}

//emu->out_logf(LOG_DEBUG,_T("ii0:%d ii1:%d"),ii[0],ii[1]);
//for(i = 0; i < 3; i++) {
//	emu->out_logf(LOG_DEBUG,_T(" y:%d y0:%d y1:%d"),i,y0[i],y1[i]);
//}

	// out of disptmg area (top or bottom padding area)
	for(i = 0; i < ii[0]; i++) {
		buf_y = - top_raster + (VRAM_BUF_HEIGHT << 1) + y0[i];
		scrn = scrn0 + scrn_offset * y0[i];
		scrnf = scrnflag[0] + scrn_offset * y0[i];
//		scrnline1p = (afterimage != 0 ? scrnline1[0] + scrnline1_offset * y0[i] : scrnline2);
		scrnline1p = scrnline1[0] + scrnline1_offset * y0[i];
		scrnline2m = ((b_p.i & BRG_COLOR_IMASK) ? 0 : 0xffffffff);

		for(y = y0[i]; y < y1[i]; ) {
			buf_y = buf_y % VRAM_BUF_HEIGHT;
			buf_y_h = (buf_y & ~1);

			if (vsync_top_raster <= y && y < vsync_bottom_raster) {
				for(x = screen_left; x < screen_right; x++) {
					scrn[x] = (color64[0][0][0] | (scrnline1p[x] & scrnline2m));
					scrnf[x] &= scrnline2m;
				}
			} else {
				for(x = screen_left; x < screen_right; x++) {
					if (hsync_left_col[buf_y_h] <= x && x < hsync_right_col[buf_y_h]) {
						scrn[x] = (color64[0][0][0] | (scrnline1p[x] & scrnline2m));
					} else {
						scrn[x] = (color16[b_p.b.b][b_p.b.r][b_p.b.g][b_p.b.i] | (scrnline1p[x] & scrnline2m));
					}
					scrnf[x] &= scrnline2m;
				}
			}
			y+=ys;
			buf_y+=ys;
			scrn += scrn_offset * ys;
			scrnf += scrn_offset * ys;
			scrnline1p += scrnline1_offset * ys;
		}
	}

	// disptmp area
	for(i = ii[0]; i < ii[1]; i++) {
		buf_y = - top_raster + (VRAM_BUF_HEIGHT << 1) + y0[i];
		scrn = scrn0 + scrn_offset * y0[i];
		scrnf = scrnflag[0] + scrn_offset * y0[i];
//		scrnline1p = (afterimage != 0 ? scrnline1[0] + scrnline1_offset * y0[i] : scrnline2);
		scrnline1p = scrnline1[0] + scrnline1_offset * y0[i];
		scrnline2m = ((b_p.i & BRG_COLOR_IMASK) ? 0 : 0xffffffff);
		l = (top_raster <= bottom_raster ? 0 : raster_per_char - ((bottom_raster - 1) % raster_per_char) - 1) | yd;

		for(y = y0[i]; y < y1[i]; ) {
			buf_y = buf_y % VRAM_BUF_HEIGHT;
			buf_y_h = (buf_y & ~1);

			// cursor line ?
			disp_cursor_line = (disp_cursor && cur_top <= y && y <= cur_bottom);

			// out of disptmg area (left padding area)
			x1 = (screen_right < left_col[buf_y_h] ? screen_right : left_col[buf_y_h]);
			for(x = screen_left; x < x1; ) {
				DISP_OUT_OF_DISPTMG;
				x++;
			}
			// out of disptmg area (right padding area)
			x1 = (screen_right < right_col[buf_y_h] ? screen_right : right_col[buf_y_h]);
			for(x = x1; x < screen_right; ) {
				DISP_OUT_OF_DISPTMG;
				x++;
			}

			// disptmg area
			buf_y_h = (buf_y & ~1);
			buf_y_h2 = cp_odd_line[buf_y_h] ? buf_y_h : buf_y;
			buf_x = left_col[buf_y_h] < 0 ? - left_col[buf_y_h] / ppch : 0 ;
			x0 = (screen_left > left_col[buf_y_h] ? screen_left : left_col[buf_y_h]);
			x1 = (screen_right < right_col[buf_y_h] ? screen_right : right_col[buf_y_h]);
			xodd = 0;

			if (NOW_S1_MODE) {
				// S1 mode
				for(x = x0; x < x1; ) {
					disp_cursor_char = (disp_cursor_line && cur_left[l] <= x && x < cur_right[l]);

					text_cram = crt_mon[buf_y_h][buf_x].color;

					// text character
					if ((text_cram & 0x50) != 0x10) {
						text_code = crt_mon[buf_y_h][buf_x].text;

						if (text_cram & 0x20) {
							// ig pattern
							ci[0].i = crt_mon[buf_y_h][buf_x].c.i;
							// reverse dot
							if (text_cram & 0x08) {
								ci[0].i = ~ci[0].i;
							}
							// cursor line ?
							if (disp_cursor_char) {
								ci[0].b.b = ((text_cram & 1) ? 0xff : 0);
								ci[0].b.r = ((text_cram & 2) ? 0xff : 0);
								ci[0].b.g = ((text_cram & 4) ? 0xff : 0);
								ci[0].b.i = 0;
							}
							// convert
							INIT_BRG_2(0,ct);
#ifdef EXPAND_ALGORITHM1
							// when width 40
							if ((text_cram & 0x80) == 0) {
								if (xodd) {
									bt0 = 0; btx0 = 1;
								} else {
									bt0 = 4; btx0 = 1;
								}
							} else {
								bt0 = 0; btx0 = 0;
							}
							for (bt = 0; bt < 8; bt++) {
								btx = (bt >> btx0) + bt0;
								iwk1.i = ((ci[0].i >> btx) & 0x01010101);
								ct[0].i |= (iwk1.i << bt);
							}
#else
							if ((text_cram & 0x80) == 0) {
								// when width 40
								if (xodd) {
									EXPAND_UINT32_LEFT(ci[0].i, ct[0].i);
								} else {
									EXPAND_UINT32_RIGHT(ci[0].i, ct[0].i);
								}
							} else {
								ct[0].i = ci[0].i;
							}
#endif
							CREATE_MASK_1(ct[0],ct_m);
						} else {
							// character
							pattern = chrfont[REG_INTERLACE_SEL][text_code][(l >> (1 - videomode)) & 0xf];
							// cursor line ?
							if (disp_cursor_char) {
								pattern = ~pattern;
							}
							// reverse dot
							if (text_cram & 0x08) {
								pattern = ~pattern;
							}
							INIT_BRG_2(0,ct);
#ifdef EXPAND_ALGORITHM1
							// when width 40
							if ((text_cram & 0x80) == 0) {
								if (xodd) {
									bt0 = 0; btx0 = 1;
								} else {
									bt0 = 4; btx0 = 1;
								}
							} else {
								bt0 = 0; btx0 = 0;
							}
							iwk1.i = crt_mon[buf_y_h][buf_x].c.i;
							for (bt = 0; bt < 8; bt++) {
								btx = (bt >> btx0) + bt0;
								if (pattern & (1 << btx)) ct[0].i |= ((iwk1.i & 0x01010101) << bt);
							}
#else
							if ((text_cram & 0x80) == 0) {
								// when width 40
								if (xodd) {
									EXPAND_UINT8_LEFT(pattern,pattern);
								} else {
									EXPAND_UINT8_RIGHT(pattern,pattern);
								}
							}
#if 1
							ct[0].i = (crt_mon[buf_y_h][buf_x].c.i & (pattern | ((uint32)pattern << 8) | ((uint32)pattern << 16) | ((uint32)pattern << 24)));
#else
							ct[0].b.b = (crt_mon[buf_y_h][buf_x].c.b.b & pattern);
							ct[0].b.r = (crt_mon[buf_y_h][buf_x].c.b.r & pattern);
							ct[0].b.g = (crt_mon[buf_y_h][buf_x].c.b.g & pattern);
							ct[0].b.i = (crt_mon[buf_y_h][buf_x].c.b.i & pattern);
#endif
#endif
							CREATE_MASK_1(ct[0],ct_m);
						}
					}
					// graphic
					attr = crt_mon[buf_y_h][buf_x].attr;
					if ((text_cram & 0x50) && (attr & DISP_PAGE_BRG_MASK)) {
						ci[0].i = (crt_mon[buf_y_h2][buf_x].g[0].i & crt_mon[buf_y_h][buf_x].brg_mask.i);

						if ((attr & SCRN_MODE_MASK) == SCRN_MODE_64COLOR_MASK) {
							// 64color
							ci[1].i = (crt_mon[buf_y_h2][buf_x].g[1].i & crt_mon[buf_y_h2][buf_x].brg_mask.i);
							INIT_BRG_2(0,cg);
#ifdef EXPAND_ALGORITHM1
							if ((crt_mon[buf_y_h][buf_x].attr & 0x02) == 0) {
								// when width 320
								if (xodd) {
									bt0 = 0; btx0 = 1;
								} else {
									bt0 = 4; btx0 = 1;
								}
							} else {
								bt0 = 0; btx0 = 0;
							}
							for (bt = 0; bt < 8; bt++) {
								btx = (bt >> btx0) + bt0;
								iwk1.i = ((ci[0].i >> btx) & 0x01010101);
								cg[0].i |= (iwk1.i << bt);
								iwk1.i = ((ci[1].i >> btx) & 0x01010101);
								cg[1].i |= (iwk1.i << bt);
							}
#else
							if ((crt_mon[buf_y_h][buf_x].attr & 0x02) == 0) {
								// when width 320
								if (xodd) {
									EXPAND_UINT32_LEFT(ci[0].i, cg[0].i);
									EXPAND_UINT32_LEFT(ci[1].i, cg[1].i);
								} else {
									EXPAND_UINT32_RIGHT(ci[0].i, cg[0].i);
									EXPAND_UINT32_RIGHT(ci[1].i, cg[1].i);
								}
							} else {
								cg[0] = ci[0];
								cg[1] = ci[1];
							}
#endif
							if (TVSUPER_DIGITAL) {
								// In digital RGB, 8 color mode
								CREATE_MASK_1(cg[0],iwk1);
								COPY_BRG_1_WITH_MASK(cg[1],cg[1],iwk1);
								cg[0].i |= cg[1].i;
								CREATE_MASK_1(cg[0],cg_m);

								if ((text_cram & 0x50) == 0x10) {
									// display graphic only
									COPY_BRG_1(cg[0],cd[0]);
								} else if ((text_cram & 0x50) == 0x40) {
									// display text and graphic (text is a top)
									COPY_BRG_1_WITH_MASK(cg[0],cd[0],ct_m);
									ADD_BRG_1(ct[0],cd[0]);
								} else {
									// display text and graphic (graphic is a top)
									COPY_BRG_1_WITH_MASK(ct[0],cd[0],cg_m);
									ADD_BRG_1(cg[0],cd[0]);
								}
								CREATE_MASK_1(cd[0],cd_m);
#ifndef PALETTE_PRESET
								// convert palette color
								INIT_BRG_1(0, cd2[0]);
								for(bt = 0; bt < 8; bt++) {
									ENCODE_PALETTE_ONE_DOT_1(bt, cd[0], cd2[0]);
								}
#endif
								// add bg color
								ADD_BRG_1_WITH_MASK(cb,cd2[0],cd_m);
								COPY_BRG_1(cd2[0],cd2[1]);

							} else {
								// analog RGB
								CREATE_MASK_2(cg,cg_m);

								if ((text_cram & 0x50) == 0x10) {
									// display graphic only
									COPY_BRG_2(cg,cd);
								} else if ((text_cram & 0x50) == 0x40) {
									// display text and graphic (text is a top)
									COPY_BRG_2_WITH_MASK(cg,cd,ct_m);
									ADD_BRG_1(ct[0],cd[0]);
									ADD_BRG_1(ct[0],cd[1]);
								} else {
									// display text and graphic (graphic is a top)
									COPY_BRG_1_WITH_MASK(ct[0],cd[0],cg_m);
									COPY_BRG_1(cd[0],cd[1]);
									ADD_BRG_2(cg,cd);
								}
								CREATE_MASK_2(cd,cd_m);
#ifndef PALETTE_PRESET
#if 1
								// non convert
								COPY_BRG_2(cd,cd2);
#else
								// convert palette color
								INIT_BRG_2(0, cd2);
								for(bt = 0; bt < 8; bt++) {
									ENCODE_PALETTE_ONE_DOT_2(bt, cd, cd2);
								}
#endif
#endif
								// add bg color
								ADD_BRG_1_WITH_MASK(cb,cd2[0],cd_m);
								ADD_BRG_1_WITH_MASK(cb,cd2[1],cd_m);
							}
						} else {
							// 
							INIT_BRG_2(0,cg);
#ifdef EXPAND_ALGORITHM1
							if ((crt_mon[buf_y_h][buf_x].attr & 0x02) == 0) {
								// when width 320
								if (xodd) {
									bt0 = 0; btx0 = 1;
								} else {
									bt0 = 4; btx0 = 1;
								}
							} else {
								bt0 = 0; btx0 = 0;
							}
							for (bt = 0; bt < 8; bt++) {
								btx = (bt >> btx0) + bt0;
								iwk1.i = ((ci[0].i >> btx) & 0x01010101);
								cg[0].i |= (iwk1.i << bt);
							}
#else
							if ((crt_mon[buf_y_h][buf_x].attr & 0x02) == 0) {
								// when width 320
								if (xodd) {
									EXPAND_UINT32_LEFT(ci[0].i, cg[0].i);
								} else {
									EXPAND_UINT32_RIGHT(ci[0].i, cg[0].i);
								}
							} else {
								cg[0] = ci[0];
							}
#endif
							CREATE_MASK_1(cg[0],cg_m);
							cg[1] = cg[0];

							if ((text_cram & 0x50) == 0x10) {
								// display graphic only
								COPY_BRG_1(cg[0],cd[0]);
							} else if ((text_cram & 0x50) == 0x40) {
								// display text and graphic (text is a top)
								COPY_BRG_1_WITH_MASK(cg[0],cd[0],ct_m);
								ADD_BRG_1(ct[0],cd[0]);
							} else {
								// display text and graphic (graphic is a top)
								COPY_BRG_1_WITH_MASK(ct[0],cd[0],cg_m);
								ADD_BRG_1(cg[0],cd[0]);
							}
							CREATE_MASK_1(cd[0],cd_m);
#ifndef PALETTE_PRESET
							if (TVSUPER_ANALOG || (attr & SCRN_MODE_MASK) == SCRN_MODE_640x400_MASK) {
								// non convert
								COPY_BRG_1(cd[0],cd2[0]);
								cd2[0].b.i = ~cd_m.b.i;
							} else {
								// convert palette color
								INIT_BRG_1(0, cd2[0]);
								for(bt = 0; bt < 8; bt++) {
									ENCODE_PALETTE_ONE_DOT_1(bt, cd[0], cd2[0]);
								}
							}
#endif
							// add bg color
							ADD_BRG_1_WITH_MASK(cb,cd2[0],cd_m);
							COPY_BRG_1(cd2[0],cd2[1]);
						}
					} else {
						// display text only
						COPY_BRG_1(ct[0],cd[0]);

						CREATE_MASK_1(cd[0],cd_m);
#ifndef PALETTE_PRESET
						if (TVSUPER_ANALOG || (attr & SCRN_MODE_MASK) == SCRN_MODE_640x400_MASK) {
							// non convert
							COPY_BRG_1(cd[0],cd2[0]);
							cd2[0].b.i = ~cd_m.b.i;
						} else {
							// convert palette color
							INIT_BRG_1(0, cd2[0]);
							for(bt = 0; bt < 8; bt++) {
								ENCODE_PALETTE_ONE_DOT_1(bt, cd[0], cd2[0]);
							}
						}
#endif
						// add bg color
						ADD_BRG_1_WITH_MASK(cb,cd2[0],cd_m);
						COPY_BRG_1(cd2[0],cd2[1]);
					}

#ifdef _DEBUG_CTRAM
	if (buf_x >= 30 && buf_y_h >= 44 && buf_y_h < 46) {
		emu->out_debugf(_T("usc txt: y:%d t:%02x c:%02x ct:%08x ctm:%08x cd:%08x"),buf_y_h,text_code,text_cram,ct[0].i,ct_m.i,cd[0].i);
	}
#endif

					// render
					if (TVSUPER_ANALOG && ((attr & SCRN_MODE_MASK) == SCRN_MODE_64COLOR_MASK)) {
						// 64color
						dt = x;
						for (bt = 0; bt < 8; bt++) {
							DECODE_BRG_2BIT_TO_PALETTE64(bt, cd2, t_p);
							dot = color64[t_p.b.b][t_p.b.r][t_p.b.g];
							scrnline2m = (t_p.i ? 0 : 0xffffffff);
							if (dt > SCREEN_WIDTH) break;
							scrn[dt] = (dot | (scrnline1p[dt] & scrnline2m));
							scrnf[dt] &= scrnline2m;
							dt++;
						}
					} else {
						// 16color
						dt = x;
						for (bt = 0; bt < 8; bt++) {
							DECODE_BRG_1BIT_TO_PALETTE16(bt, cd2, t_p);
							dot = color16[t_p.b.b][t_p.b.r][t_p.b.g][t_p.b.i];
							scrnline2m = (t_p.i ? 0 : 0xffffffff);
							if (dt > SCREEN_WIDTH) break;
							scrn[dt] = (dot | (scrnline1p[dt] & scrnline2m));
							scrnf[dt] &= scrnline2m;
							dt++;
						}
					}
					x += 8;
					buf_x++;
					xodd = (1 - xodd);
				}
			} else {
				// L3 mode
				for(x = x0; x < x1; ) {
					disp_cursor_char = (disp_cursor_line && cur_left[l] <= x && x < cur_right[l]);

					text_code = crt_mon[buf_y_h][buf_x].text;
					text_cram = crt_mon[buf_y_h][buf_x].color;
					c_f = (text_cram & 0x07) | 0x08;
					DECODE_TO_PALETTE16_NUM(c_f, t_p);
					c_b = bg_color;

					switch(text_cram & 0x30) {
					case 0x30:
					case 0x20:
						// ig
						// disp ig pattern
						DISP_L3_IG_PATTERN(text_code);
						break;
					case 0x10:
						// graphic
						if (text_cram & 8) text_code = ~text_code;
						if (disp_cursor_char) text_code = ~text_code;
						if (text_cram & 0x40) {
							// hireso graphic
							DISP_L3_GRAPHIC_HIRESO(text_code);
						} else {
							// normal graphic
							DISP_L3_GRAPHIC_NORMAL(text_code);
						}
						break;
					default:
						// character
						DISP_L3_TEXT_CHAR(text_code);
						break;
					}
					x += ppch;
					buf_x++;
				}
			}

			y+=ys;
			buf_y+=ys;
			scrn += scrn_offset * ys;
			scrnf += scrn_offset * ys;
			scrnline1p += scrnline1_offset * ys;
			l = ((l + ys) % raster_per_char);
		}
	}

	// clear non drawing line
	if (scanline) {
		yd = 1 - yd;
		scrn = scrn0 + scrn_offset * (screen_top | yd);
		scrnf = scrnflag[0] + scrn_offset * (screen_top | yd);
		scrnline1p = scrnline1[0] + scrnline1_offset * (screen_top | yd);
		for(y = (screen_top | yd); y < screen_bottom; ) {
			for(x = screen_left; x < screen_right; x++) {
				scrn[x] = (color64[0][0][0] | (scrnline1p[x] & scrnline2m));
			}
			y+=ys;
			scrn += scrn_offset * ys;
			scrnf += scrn_offset * ys;
			scrnline1p += scrnline1_offset * ys;
		}
	}
}

#define SET_CRT_MON(row, name, value) \
	crt_mon[row][crt_mon_col].name = (value);

#define ADD_CRT_MON(row, name, value) \
	crt_mon[row][crt_mon_col].name |= (value);

#define GET_CRT_MON(row, name) \
	crt_mon[row][crt_mon_col].name

#define P_SET_CRT_MON(ptr, name, value) \
	ptr->name = (value);

#define P_ADD_CRT_MON(ptr, name, value) \
	ptr->name |= (value);

#define P_GET_CRT_MON(ptr, name) \
	ptr->name

void DISPLAY::update_vram(int v, int clock)
{
	static int addr;
	static uint8 text,color,val,valr;
	static int crt_mon_row2;
	static int vtaddr;
	static int vtadiv;
	static int vgaddr;
	static int vgaddr2;
	static int vgadiv;
	static t_crt_mon *p_crt_mon, *p_crt_mon2;

	if (((*crtc_vt_count) << 1) != crt_mon_row) {
		if (*crtc_vt_count == 0) {
			// disptmg on

			// calculate start addr from crtc
			crtc_start_addr = (((regs[12] << 8) | regs[13]) + disptmg_skew - config.disptmg_skew) & 0x3fff;
			if (NOW_L3_MODE) {
				if (crtc_start_addr < L3_ADDR_VRAM_START) {
					crtc_start_addr += (L3_ADDR_VRAM_END - L3_ADDR_VRAM_START);
				}
			}

			// pixel of height
			top_raster = now_raster * 2;
			top_raster += (SCREEN_HEIGHT - 488) / 2;
			bottom_raster = top_raster + (regs[6] * raster_per_char);

			vsync_top_raster = top_raster + (regs[7] * raster_per_char);
			vsync_bottom_raster = vsync_top_raster + (((regs[3] & 0xf0) ? (regs[3] >> 4) : 16) << 1);

//			top_raster = (top_raster + lines_per_2frames) % lines_per_2frames;
//			bottom_raster = (bottom_raster + lines_per_2frames) % lines_per_2frames;

//			vsync_top_raster = (vsync_top_raster + lines_per_2frames) % lines_per_2frames;
//			vsync_bottom_raster = (vsync_bottom_raster + lines_per_2frames) % lines_per_2frames;

			// synchronize horizontal total
			if (h_total_min[crtc_chr_clocks] <= regs[0] && regs[0] <= h_total_max[crtc_chr_clocks]) {
				left_diff = 0;
			}
			// top most page ? (L3 only)
//			if (NOW_L3_MODE) {
//				crt_mon_page = ((keepimage == 1 && start_addr < 0x0800) || (keepimage == 2 && start_addr >= 0x0800)) ? 0x80 : 0;
//			}
		}

		crt_mon_row = ((*crtc_vt_count) << 1);
		crt_mon_col = 0;

		if (NOW_S1_MODE) {
			// S1 mode
			crt_mon_raster = (crt_mon_row % raster_per_char);

			crt_mon_s1vtaddr = crtc_start_addr + (crt_mon_row / raster_per_char) * width;
			crt_mon_s1vgaddr = crtc_start_addr + (crt_mon_row / raster_per_char) * gwidth;
			if (SCRN_MODE_IS_640x400) {
				crt_mon_s1vgaddr += (crt_mon_raster & 0xf) * 0x800;
			} else {
				crt_mon_s1vgaddr += ((crt_mon_raster >> 1) & 0x7) * (0x400 << gwidth_sel); 
			}
			// cursor
			cursor_addr = (((regs[14] << 8) | regs[15]) & 0x3fff) + disptmg_skew - config.disptmg_skew;
			cursor_addr >>= (1 - width_sel); 

		} else {
			// L3 mode
			crt_mon_raster = (crt_mon_row % raster_per_char);

			crt_mon_l3vaddr = crtc_start_addr + (crt_mon_row / raster_per_char) * width;
			crt_mon_l3vaddr += hireso * ((crt_mon_raster >> 1) & 0x7) * (0x400 << width_sel);
			// cursor
			cursor_addr = (((regs[14] << 8) | regs[15]) & 0x3fff) + disptmg_skew - config.disptmg_skew;
			if (cursor_addr < 0x400) cursor_addr += 0x4000;

		}
		// pixel of width
		if (crt_mon_row < VRAM_BUF_HEIGHT) {
			left_col[crt_mon_row] = regs[0]-regs[2]-(regs[3] & 0x0f)-(crtc_chr_clocks ? 14 : 5) + disptmg_skew - config.disptmg_skew;
			left_col[crt_mon_row] *= crtc_ppch;
			left_col[crt_mon_row] += ((SCREEN_WIDTH - 768) / 2);
			left_col[crt_mon_row] += left_diff;

			right_col[crt_mon_row] = left_col[crt_mon_row] + crtc_ppch * regs[1];

			hsync_left_col[crt_mon_row] = left_col[crt_mon_row] + regs[2] * crtc_ppch;
			hsync_right_col[crt_mon_row] = hsync_left_col[crt_mon_row] + (regs[3] & 0x0f) * crtc_ppch;
		}
		// diff horizontal total display and crtc
		if (regs[0] < h_total_min[crtc_chr_clocks] || h_total_max[crtc_chr_clocks] < regs[0]) {
			left_diff += (regs[0] + 1 - (1 << (crtc_chr_clocks + 6))) * crtc_ppch;

			if (left_diff >= SCREEN_WIDTH) left_diff = -SCREEN_WIDTH + 1;
			else if (left_diff <= -SCREEN_WIDTH) left_diff = SCREEN_WIDTH - 1;
		}

		now_raster++;
		if (now_raster >= lines_per_2frames) now_raster = 0;
	}

	// copy vram data to buffer
	if (crt_mon_row < VRAM_BUF_HEIGHT && *crtc_vt_count < *crtc_vt_disp) {
		cp_odd_line[crt_mon_row] = 1;

		if (NOW_S1_MODE) {
			// S1 mode
			crt_mon_row2 = crt_mon_row + 1;
			if (crt_mon_row2 >= VRAM_BUF_HEIGHT) crt_mon_row2 = 0;
			// add page number 
			vtaddr = (crt_mon_s1vtaddr + (width_sel ? 0 : page_no) * (S1_VTRAM_SIZE >> 1));
			vtadiv = 1;
			vgaddr = (crt_mon_s1vgaddr + ((gwidth_sel || SCRN_MODE_IS_64COLOR) ? 0 : gpage_no) * (S1_VGRAM_SIZE_ONE >> 1));
			vgaddr2 = -1;
			vgadiv = 1;

			if (SCRN_MODE_IS_64COLOR) {
				// 64color mode
				vgaddr2 = vgaddr + (S1_VGRAM_SIZE_ONE >> 1);
			} else if (SCRN_MODE_IS_640x400) {
				// 640x400 mode
				vgaddr2 = vgaddr + 0x800;
				cp_odd_line[crt_mon_row] = 0;
			}
			for(int i=0; i < VRAM_BUF_WIDTH; i++) {
				p_crt_mon  = &crt_mon[crt_mon_row][crt_mon_col];
				p_crt_mon2 = &crt_mon[crt_mon_row2][crt_mon_col];

				if (vtaddr >= S1_VTRAM_SIZE) {
					vtaddr -= S1_VTRAM_SIZE;
				}
				if (SCRN_MODE_IS_640x400) {
					if (vgaddr >= S1_VGRAM_SIZE_TWO) {
						vgaddr -= S1_VGRAM_SIZE_TWO;
					}
				} else {
					if (vgaddr >= S1_VGRAM_SIZE_ONE) {
						vgaddr -= S1_VGRAM_SIZE_ONE;
					}
				}
				if (crt_mon_col < regs[1]) {
					// text vram
					text = s1vtram[vtaddr];
					P_SET_CRT_MON(p_crt_mon, text, text);
					// text color
					// bit 0-6: color ram  bit 7:width 0:40 1:80
					color = ((s1vcram[vtaddr] & 0x7f) | (width_sel << 7));
#ifdef _DEBUG_CVRAM
	if (vtaddr >= 0x0 && vtaddr < 0x28 && (crt_mon_row & 7) == 0) {
		emu->out_debugf(_T("up_v txt: y:%d %04x t:%02x c:%02x"),crt_mon_row,vtaddr,text,color);
	}
#endif
					P_SET_CRT_MON(p_crt_mon, color, color);
					if (color & 0x20) {
						// ig
						addr = (text << 3)+(crt_mon_raster >> 1);
						P_SET_CRT_MON(p_crt_mon, c.b.b, ig_ram[addr]);
						P_SET_CRT_MON(p_crt_mon, c.b.r, ig_ram[addr+IGRAM_SIZE]);
						P_SET_CRT_MON(p_crt_mon, c.b.g, ig_ram[addr+IGRAM_SIZE+IGRAM_SIZE]);
#ifdef PALETTE_PRESET
						// convert palette color
						iwk3.i = 0;
						for(int i=0; i<8; i++) {
							ENCODE_PALETTE_ONE_DOT_I(i, GET_CRT_MON(crt_mon_row, c), iwk2);
							iwk3.i |= (iwk2.i << i);
						}
						P_SET_CRT_MON(p_crt_mon, c.i, iwk3.i);
#else
						P_SET_CRT_MON(p_crt_mon, c.b.i, 0);
#endif
					} else {
						// convert palette color
#ifdef PALETTE_PRESET
						P_SET_CRT_MON(p_crt_mon, c.i, palette_brg[color & 1][(color & 2) >> 1][(color & 4) >> 2].i);
#else
						P_SET_CRT_MON(p_crt_mon, c.b.b, (color & 1) ? 0xff : 0);
						P_SET_CRT_MON(p_crt_mon, c.b.r, (color & 2) ? 0xff : 0);
						P_SET_CRT_MON(p_crt_mon, c.b.g, (color & 4) ? 0xff : 0);
						P_SET_CRT_MON(p_crt_mon, c.b.i, 0);
#endif
					}
					// cursor
					if (vtaddr == cursor_addr && crt_mon_raster == 0) {
						if (crt_mon_cur_col < 0) crt_mon_cur_col = crt_mon_col;
						crt_mon_cur_row = crt_mon_row;
					}

					// graphic vram
					if (SCRN_MODE_IS_640x400) {
						// 640x400 mono
						// even line
						val = (vgaddr < S1_VGRAM_SIZE_ONE ? s1vgrram[vgaddr] : s1vggram[vgaddr-S1_VGRAM_SIZE_ONE]);
						P_SET_CRT_MON(p_crt_mon, g[0].b.r, palette_brg[0][1][0].b.r * val);
						P_SET_CRT_MON(p_crt_mon, g[0].b.g, palette_brg[0][1][0].b.g * val);
						P_SET_CRT_MON(p_crt_mon, g[0].b.b, palette_brg[0][1][0].b.b * val);
						P_SET_CRT_MON(p_crt_mon, g[0].b.i, palette_brg[0][1][0].b.i * val);
						valr = ~val;
						P_ADD_CRT_MON(p_crt_mon, g[0].b.r, palette_brg[0][0][0].b.r * valr);
						P_ADD_CRT_MON(p_crt_mon, g[0].b.g, palette_brg[0][0][0].b.g * valr);
						P_ADD_CRT_MON(p_crt_mon, g[0].b.b, palette_brg[0][0][0].b.b * valr);
						P_ADD_CRT_MON(p_crt_mon, g[0].b.i, palette_brg[0][0][0].b.i * valr);

						// odd line
						val = (vgaddr2 < S1_VGRAM_SIZE_ONE ? s1vgrram[vgaddr2] : s1vggram[vgaddr2-S1_VGRAM_SIZE_ONE]);
						P_SET_CRT_MON(p_crt_mon2, g[0].b.r, palette_brg[0][0][1].b.r * val);
						P_SET_CRT_MON(p_crt_mon2, g[0].b.g, palette_brg[0][0][1].b.g * val);
						P_SET_CRT_MON(p_crt_mon2, g[0].b.b, palette_brg[0][0][1].b.b * val);
						P_SET_CRT_MON(p_crt_mon2, g[0].b.i, palette_brg[0][0][1].b.i * val);
						valr = ~val;
						P_ADD_CRT_MON(p_crt_mon2, g[0].b.r, palette_brg[0][0][0].b.r * valr);
						P_ADD_CRT_MON(p_crt_mon2, g[0].b.g, palette_brg[0][0][0].b.g * valr);
						P_ADD_CRT_MON(p_crt_mon2, g[0].b.b, palette_brg[0][0][0].b.b * valr);
						P_ADD_CRT_MON(p_crt_mon2, g[0].b.i, palette_brg[0][0][0].b.i * valr);

						// attribute
						P_SET_CRT_MON(p_crt_mon, brg_mask.i, 0xffffffff);
					} else {
						// normal
						P_SET_CRT_MON(p_crt_mon, g[0].b.r, s1vgrram[vgaddr]);
						P_SET_CRT_MON(p_crt_mon, g[0].b.g, s1vggram[vgaddr]);
						P_SET_CRT_MON(p_crt_mon, g[0].b.b, s1vgbram[vgaddr]);
						P_SET_CRT_MON(p_crt_mon, g[0].b.i, 0);
#ifdef PALETTE_PRESET
						// convert palette color
						iwk3.i = 0;
						for(int i=0; i<8; i++) {
							ENCODE_PALETTE_ONE_DOT_I(i, GET_CRT_MON(crt_mon_row, g[0]), iwk2);
							iwk3.i |= (iwk2.i << i);
						}
						P_SET_CRT_MON(p_crt_mon, g[0].i, iwk3.i);
#endif
						if (vgaddr2 >= 0) {
							// 320x200 64color
							P_SET_CRT_MON(p_crt_mon, g[1].b.r, s1vgrram[vgaddr2]);
							P_SET_CRT_MON(p_crt_mon, g[1].b.g, s1vggram[vgaddr2]);
							P_SET_CRT_MON(p_crt_mon, g[1].b.b, s1vgbram[vgaddr2]);
							P_SET_CRT_MON(p_crt_mon, g[1].b.i, 0);
#ifdef PALETTE_PRESET
							// convert palette color
							iwk3.i = 0;
							for(int i=0; i<8; i++) {
								ENCODE_PALETTE_ONE_DOT_I(i, GET_CRT_MON(crt_mon_row, g[1]), iwk2);
								iwk3.i |= (iwk2.i << i);
							}
							P_SET_CRT_MON(p_crt_mon, g[1].i, iwk3.i);
#endif
						}
						// attribute
						P_SET_CRT_MON(p_crt_mon, brg_mask, disp_page_brg_mask);
					}
					//   bit0-1:scrn_mode  bit5-7:disp graphic plane b g r
					P_SET_CRT_MON(p_crt_mon, attr, ((DISP_PAGE_GRAPHIC_BRG) | (SCRN_MODE_GRAPHIC)));

				} else if (crt_mon_col < VRAM_BUF_WIDTH) {
					P_SET_CRT_MON(p_crt_mon, text, 0);
					P_SET_CRT_MON(p_crt_mon, color, 0);
					P_SET_CRT_MON(p_crt_mon, g[0].i, 0);
					P_SET_CRT_MON(p_crt_mon, g[1].i, 0);
					P_SET_CRT_MON(p_crt_mon, attr, 0);
				}
				crt_mon_col++;
				vtadiv = 1 - vtadiv;
				vtaddr+=(width_sel ? 1 : vtadiv);
				vgadiv = 1 - vgadiv;
				vgaddr+=(gwidth_sel ? 1 : vgadiv);
				if (vgaddr2 >= 0) vgaddr2+=(gwidth_sel ? 1 : vgadiv);
			}
			crt_mon_s1vtaddr+=vtaddr;
			crt_mon_s1vgaddr+=vgaddr;
		} else {
			// L3 mode
			for(int i=0; i < (clock << width_sel); i++) {
				p_crt_mon = &crt_mon[crt_mon_row][crt_mon_col];

				if (crt_mon_l3vaddr >= L3_ADDR_VRAM_END) {
					crt_mon_l3vaddr -= (L3_ADDR_VRAM_END - L3_ADDR_VRAM_START);
				}
				if (crt_mon_col < width) {
					P_SET_CRT_MON(p_crt_mon, text, l3vram[crt_mon_l3vaddr]);
					// bit 0-5: color ram   bit 6: hireso
					P_SET_CRT_MON(p_crt_mon, color, (color_ram[crt_mon_l3vaddr] | (hireso << 6) | crt_mon_page) & (ig_mask ? 0xdf : 0xff));
					// cursor
					if (crt_mon_l3vaddr == cursor_addr && crt_mon_raster == 0) {
						crt_mon_cur_col = crt_mon_col;
						crt_mon_cur_row = crt_mon_row;
					}
				} else if (crt_mon_col < VRAM_BUF_WIDTH) {
					P_SET_CRT_MON(p_crt_mon, text, 0);
					P_SET_CRT_MON(p_crt_mon, color, 0);
				}
				crt_mon_col++;
				crt_mon_l3vaddr++;
			}
		}
	}
}

// ----------------------------------------------------------------------------
void DISPLAY::set_palette_brg()
{
	for(int i=0; i<8; i++) {
		set_palette_brg(i);
	}
}

void DISPLAY::set_palette_brg(int num)
{
	int pb =  (num & 1);
	int pr = ((num & 2) >> 1);
	int pg = ((num & 4) >> 2);
	palette_brg[pb][pr][pg].b.b = (REG_PALETTE[num] & 1 ? 1 : 0);
	palette_brg[pb][pr][pg].b.r = (REG_PALETTE[num] & 2 ? 1 : 0);
	palette_brg[pb][pr][pg].b.g = (REG_PALETTE[num] & 4 ? 1 : 0);
	palette_brg[pb][pr][pg].b.i = (REG_PALETTE[num] & 8 ? 1 : 0);
}

void DISPLAY::set_disp_page()
{
	page_no = (DISP_PAGE_TEXT) ? 1 : 0;
	gpage_no = (DISP_PAGE_GRAPHIC) ? 1 : 0;
	disp_page_brg_mask.b.b = (DISP_PAGE_GRAPHIC_B ? 0xff : 0);
	disp_page_brg_mask.b.r = (DISP_PAGE_GRAPHIC_R ? 0xff : 0);
	disp_page_brg_mask.b.g = (DISP_PAGE_GRAPHIC_G ? 0xff : 0);
	disp_page_brg_mask.b.i = 0xff;
}

void DISPLAY::set_s1_screen_mode()
{
	// for text
	width_sel = SCRN_MODE_TEXT_WIDTH ? 1 : 0;	// 0:w.40  1:w.80
	width = width_sel ? 80 : 40;

//			for(int i=0; i<8; i++) {
//				ppd[i]=width_sel ? 1 : 2;	// pixel per 1dot
//				dof[i]=(0x80 >> i);
//			}
//			// pixel per 1char
//			ppch = width_sel ? 8 : 16;

	// for graphic
	gwidth_sel = SCRN_MODE_GRAPHIC_WIDTH ? 1 : 0;	// 640 or 320
	gwidth = SCRN_MODE_GRAPHIC_WIDTH ? 80 : 40;	// 640 or 320
}

// ----------------------------------------------------------------------------
// event handler
// ----------------------------------------------------------------------------

void DISPLAY::event_frame()
{
//	cblink = (cblink + (1 << config.cpu_power)) & 0x1f;
//	sw = 1 - sw;
}

// ----------------------------------------------------------------------------

void DISPLAY::save_state(FILEIO *fio)
{
	struct vm_state_st vm_state;

	//
	vm_state_ident.version = 0x41;
	vm_state_ident.size = sizeof(vm_state_ident) + sizeof(vm_state);

	// copy values
	memset(&vm_state, 0, sizeof(vm_state));
	vm_state.crtc_addr = crtc_addr;
	vm_state.ig_enable = REG_IGENABLE;
	vm_state.ig_modereg = REG_IGMODE;
	vm_state.ig_mask = ig_mask ? 1 : 0;

	vm_state.sys_mode = REG_SYS_MODE;
	vm_state.mode_sel = REG_MODE_SEL;

	vm_state.chr_clocks = (crtc_chr_clocks & 0xff);
	vm_state.now_cpu_halt = now_cpu_halt;
	vm_state.interlace_sel = REG_INTERLACE_SEL;
	vm_state.tvsuper = REG_TVSUPER;

	//
	vm_state.bmsk_color = REG_BMSK_COLOR;
	vm_state.active_plane = REG_ACTIVE_PLANE;
	vm_state.dummy1 = 0;
	vm_state.disp_page = REG_DISP_PAGE;
	vm_state.scrn_mode = REG_SCRN_MODE;
	memcpy(vm_state.gcolor, REG_GCOLOR, sizeof(vm_state.gcolor));
	memcpy(vm_state.palette, REG_PALETTE, sizeof(vm_state.palette));

	fio->Fwrite(&vm_state_ident, sizeof(vm_state_ident), 1);
	fio->Fwrite(&vm_state, sizeof(vm_state), 1);
}

void DISPLAY::load_state(FILEIO *fio)
{
	vm_state_ident_t vm_state_i;
	struct vm_state_st vm_state;
	uint8 mode_sel = 0;

	READ_STATE_CHUNK(fio, vm_state_i, vm_state);

	// require CRTC register
	for(crtc_addr = 0; crtc_addr < 16; crtc_addr++) {
		write_io8(0xffc7, regs[crtc_addr]);
	}

	crtc_addr = vm_state.crtc_addr;
	REG_IGENABLE = vm_state.ig_enable;
	ig_mask = vm_state.ig_mask ? true : false;

	if (vm_state_i.version >= 2) {
		crtc_chr_clocks = vm_state.chr_clocks;
		crtc_ppch = (crtc_chr_clocks ? 8 : 16);
	}

	if (vm_state_i.version >= 0x41) {
		REG_SYS_MODE = vm_state.sys_mode;
		mode_sel = vm_state.mode_sel;

		REG_IGMODE = vm_state.ig_modereg;
		now_cpu_halt = vm_state.now_cpu_halt;
		REG_INTERLACE_SEL = vm_state.interlace_sel;
		REG_TVSUPER = vm_state.tvsuper;

		REG_BMSK_COLOR = vm_state.bmsk_color;
		REG_ACTIVE_PLANE = vm_state.active_plane;
		REG_DISP_PAGE = vm_state.disp_page;
		REG_SCRN_MODE = vm_state.scrn_mode;
		memcpy(REG_GCOLOR, vm_state.gcolor, sizeof(REG_GCOLOR));
		memcpy(REG_PALETTE, vm_state.palette, sizeof(REG_PALETTE));

		SIG_BMSK_DW = (BMSK_COLOR_BMSK ? (SIG_BMSK_DW | 2) : (SIG_BMSK_DW & ~2));
		SIG_BMSK_DW = (ACTIVE_PLANE_DW ? (SIG_BMSK_DW | 1) : (SIG_BMSK_DW & ~1));

		set_disp_page();

		if (NOW_S1_MODE) {
			set_s1_screen_mode();
		}
		set_palette_brg();

		write_io8(0xffd0, mode_sel);
	}

	// clear
	crt_mon_row = 9999;
	crt_mon_col = 0;

}

